"""
Support for the Mercurial SCM
"""

import logging

import salt.utils.data
import salt.utils.path
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if hg is installed
    """
    if salt.utils.path.which("hg") is None:
        return (False, "The hg execution module cannot be loaded: hg unavailable.")
    else:
        return True


def _ssh_flag(identity_path):
    return ["--ssh", f"ssh -i {identity_path}"]


def revision(cwd, rev="tip", short=False, user=None):
    """
    Returns the long hash of a given identifier (hash, branch, tag, HEAD, etc)

    cwd
        The path to the Mercurial repository

    rev: tip
        The revision

    short: False
        Return an abbreviated commit hash

    user : None
        Run hg as a user other than what the minion runs as

    CLI Example:

    .. code-block:: bash

        salt '*' hg.revision /path/to/repo mybranch
    """
    cmd = ["hg", "id", "-i", "--debug" if not short else "", "-r", f"{rev}"]

    result = __salt__["cmd.run_all"](cmd, cwd=cwd, runas=user, python_shell=False)

    if result["retcode"] == 0:
        return result["stdout"]
    else:
        return ""


def describe(cwd, rev="tip", user=None):
    """
    Mimic git describe and return an identifier for the given revision

    cwd
        The path to the Mercurial repository

    rev: tip
        The path to the archive tarball

    user : None
        Run hg as a user other than what the minion runs as

    CLI Example:

    .. code-block:: bash

        salt '*' hg.describe /path/to/repo
    """
    cmd = [
        "hg",
        "log",
        "-r",
        f"{rev}",
        "--template",
        "'{{latesttag}}-{{latesttagdistance}}-{{node|short}}'",
    ]
    desc = __salt__["cmd.run_stdout"](cmd, cwd=cwd, runas=user, python_shell=False)

    return desc or revision(cwd, rev, short=True)


def archive(cwd, output, rev="tip", fmt=None, prefix=None, user=None):
    """
    Export a tarball from the repository

    cwd
        The path to the Mercurial repository

    output
        The path to the archive tarball

    rev: tip
        The revision to create an archive from

    fmt: None
        Format of the resulting archive. Mercurial supports: tar,
        tbz2, tgz, zip, uzip, and files formats.

    prefix : None
        Prepend <prefix>/ to every filename in the archive

    user : None
        Run hg as a user other than what the minion runs as

    If ``prefix`` is not specified it defaults to the basename of the repo
    directory.

    CLI Example:

    .. code-block:: bash

        salt '*' hg.archive /path/to/repo output=/tmp/archive.tgz fmt=tgz
    """
    cmd = [
        "hg",
        "archive",
        f"{output}",
        "--rev",
        f"{rev}",
    ]
    if fmt:
        cmd.append("--type")
        cmd.append(f"{fmt}")
    if prefix:
        cmd.append("--prefix")
        cmd.append(f'"{prefix}"')
    return __salt__["cmd.run"](cmd, cwd=cwd, runas=user, python_shell=False)


def pull(cwd, opts=None, user=None, identity=None, repository=None):
    """
    Perform a pull on the given repository

    cwd
        The path to the Mercurial repository

    repository : None
        Perform pull from the repository different from .hg/hgrc:[paths]:default

    opts : None
        Any additional options to add to the command line

    user : None
        Run hg as a user other than what the minion runs as

    identity : None
        Private SSH key on the minion server for authentication (ssh://)

        .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' hg.pull /path/to/repo opts=-u
    """
    cmd = ["hg", "pull"]
    if identity:
        cmd.extend(_ssh_flag(identity))
    if opts:
        for opt in opts.split():
            cmd.append(opt)
    if repository is not None:
        cmd.append(repository)

    ret = __salt__["cmd.run_all"](cmd, cwd=cwd, runas=user, python_shell=False)
    if ret["retcode"] != 0:
        raise CommandExecutionError(
            "Hg command failed: {}".format(ret.get("stderr", ret["stdout"]))
        )

    return ret["stdout"]


def update(cwd, rev, force=False, user=None):
    """
    Update to a given revision

    cwd
        The path to the Mercurial repository

    rev
        The revision to update to

    force : False
        Force an update

    user : None
        Run hg as a user other than what the minion runs as

    CLI Example:

    .. code-block:: bash

        salt devserver1 hg.update /path/to/repo somebranch
    """
    cmd = ["hg", "update", f"{rev}"]
    if force:
        cmd.append("-C")

    ret = __salt__["cmd.run_all"](cmd, cwd=cwd, runas=user, python_shell=False)
    if ret["retcode"] != 0:
        raise CommandExecutionError(
            "Hg command failed: {}".format(ret.get("stderr", ret["stdout"]))
        )

    return ret["stdout"]


def clone(cwd, repository, opts=None, user=None, identity=None):
    """
    Clone a new repository

    cwd
        The path to the Mercurial repository

    repository
        The hg URI of the repository

    opts : None
        Any additional options to add to the command line

    user : None
        Run hg as a user other than what the minion runs as

    identity : None
        Private SSH key on the minion server for authentication (ssh://)

        .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' hg.clone /path/to/repo https://bitbucket.org/birkenfeld/sphinx
    """
    cmd = ["hg", "clone", f"{repository}", f"{cwd}"]
    if opts:
        for opt in opts.split():
            cmd.append(f"{opt}")
    if identity:
        cmd.extend(_ssh_flag(identity))

    ret = __salt__["cmd.run_all"](cmd, runas=user, python_shell=False)
    if ret["retcode"] != 0:
        raise CommandExecutionError(
            "Hg command failed: {}".format(ret.get("stderr", ret["stdout"]))
        )

    return ret["stdout"]


def status(cwd, opts=None, user=None):
    """
    Show changed files of the given repository

    cwd
        The path to the Mercurial repository

    opts : None
        Any additional options to add to the command line

    user : None
        Run hg as a user other than what the minion runs as

    CLI Example:

    .. code-block:: bash

        salt '*' hg.status /path/to/repo
    """

    def _status(cwd):
        cmd = ["hg", "status"]
        if opts:
            for opt in opts.split():
                cmd.append(f"{opt}")
        out = __salt__["cmd.run_stdout"](cmd, cwd=cwd, runas=user, python_shell=False)
        types = {
            "M": "modified",
            "A": "added",
            "R": "removed",
            "C": "clean",
            "!": "missing",
            "?": "not tracked",
            "I": "ignored",
            " ": "origin of the previous file",
        }
        ret = {}
        for line in out.splitlines():
            t, f = types[line[0]], line[2:]
            if t not in ret:
                ret[t] = []
            ret[t].append(f)
        return ret

    if salt.utils.data.is_iter(cwd):
        return {cwd: _status(cwd) for cwd in cwd}
    else:
        return _status(cwd)



import pytest

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


def test_state_with_import(salt_ssh_cli, state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run("state.sls", "test")
    assert ret.returncode == 0
    assert ret.data


@pytest.mark.parametrize(
    "ssh_cmd",
    [
        "state.sls",
        "state.highstate",
        "state.apply",
        "state.show_top",
        "state.show_highstate",
        "state.show_low_sls",
        "state.show_lowstate",
        "state.sls_id",
        "state.show_sls",
        "state.top",
    ],
)
def test_state_with_import_dir(salt_ssh_cli, state_tree_dir, ssh_cmd):
    """
    verify salt-ssh can use imported map files in states
    when the map files are in another directory outside of
    sls files importing them.
    """
    if ssh_cmd in ("state.sls", "state.show_low_sls", "state.show_sls"):
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd, "testdir")
    elif ssh_cmd == "state.top":
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd, "top.sls")
    elif ssh_cmd == "state.sls_id":
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd, "Ok with def", "testdir")
    else:
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd)
    assert ret.returncode == 0
    if ssh_cmd == "state.show_top":
        assert ret.data == {"base": ["testdir", "master_tops_test"]} or {
            "base": ["testdir"]
        }
    elif ssh_cmd in ("state.show_highstate", "state.show_sls"):
        assert ret.data == {
            "Ok with def": {
                "__sls__": "testdir",
                "__env__": "base",
                "test": ["succeed_without_changes", {"order": 10000}],
            }
        }
    elif ssh_cmd in ("state.show_low_sls", "state.show_lowstate", "state.show_sls"):
        assert ret.data == [
            {
                "state": "test",
                "name": "Ok with def",
                "__sls__": "testdir",
                "__env__": "base",
                "__id__": "Ok with def",
                "order": 10000,
                "fun": "succeed_without_changes",
            }
        ]
    else:
        assert ret.data["test_|-Ok with def_|-Ok with def_|-succeed_without_changes"][
            "result"
        ]
    assert ret.data


def test_state_with_import_from_dir(salt_ssh_cli, nested_state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run(
        "--extra-filerefs=salt://foo/map.jinja", "state.apply", "foo"
    )
    assert ret.returncode == 0
    assert ret.data


def test_state_low(salt_ssh_cli):
    """
    test state.low with salt-ssh
    """
    ret = salt_ssh_cli.run(
        "state.low", '{"state": "cmd", "fun": "run", "name": "echo blah"}'
    )
    assert ret.data["cmd_|-echo blah_|-echo blah_|-run"]["changes"]["stdout"] == "blah"


def test_state_high(salt_ssh_cli):
    """
    test state.high with salt-ssh
    """
    ret = salt_ssh_cli.run("state.high", '{"echo blah": {"cmd": ["run"]}}')
    assert ret.data["cmd_|-echo blah_|-echo blah_|-run"]["changes"]["stdout"] == "blah"


def test_state_test(salt_ssh_cli, state_tree):
    ret = salt_ssh_cli.run("state.test", "test")
    assert ret.returncode == 0
    assert ret.data
    assert (
        ret.data["test_|-Ok with def_|-Ok with def_|-succeed_with_changes"]["result"]
        is None
    )



"""
Management of OpenStack Keystone Projects
=========================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.keystoneng` for setup instructions

Example States

.. code-block:: yaml

    create project:
      keystone_project.present:
        - name: project1

    delete project:
      keystone_project.absent:
        - name: project1

    create project with optional params:
      keystone_project.present:
        - name: project1
        - domain: domain1
        - enabled: False
        - description: 'my project'
"""

__virtualname__ = "keystone_project"


def __virtual__():
    if "keystoneng.project_get" in __salt__:
        return __virtualname__
    return (
        False,
        "The keystoneng execution module failed to load: shade python module is not"
        " available",
    )


def _common(name, kwargs):
    """
    Returns: None if project wasn't found, otherwise a group object
    """
    search_kwargs = {"name": name}
    if "domain" in kwargs:
        domain = __salt__["keystoneng.get_entity"]("domain", name=kwargs.pop("domain"))
        domain_id = domain.id if hasattr(domain, "id") else domain
        search_kwargs["domain_id"] = domain_id
        kwargs["domain_id"] = domain_id

    return __salt__["keystoneng.project_get"](**search_kwargs)


def present(name, auth=None, **kwargs):
    """
    Ensure a project exists and is up-to-date

    name
        Name of the project

    domain
        The name or id of the domain

    description
        An arbitrary description of the project
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_clouds"](auth)

    kwargs["name"] = name
    project = _common(name, kwargs)

    if project is None:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Project will be created."
            return ret

        project = __salt__["keystoneng.project_create"](**kwargs)
        ret["changes"] = project
        ret["comment"] = "Created project"
        return ret

    changes = __salt__["keystoneng.compare_changes"](project, **kwargs)
    if changes:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = changes
            ret["comment"] = "Project will be updated."
            return ret

        __salt__["keystoneng.project_update"](**kwargs)
        ret["changes"].update(changes)
        ret["comment"] = "Updated project"

    return ret


def absent(name, auth=None, **kwargs):
    """
    Ensure a project does not exists

    name
        Name of the project

    domain
        The name or id of the domain
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_clouds"](auth)

    kwargs["name"] = name
    project = _common(name, kwargs)

    if project:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"id": project.id}
            ret["comment"] = "Project will be deleted."
            return ret

        __salt__["keystoneng.project_delete"](name=project)
        ret["changes"]["id"] = project.id
        ret["comment"] = "Deleted project"

    return ret



"""
    :codeauthor: Gary T. Giesen <ggiesen@giesen.me>
"""

import pytest

import salt.pillar.netbox as netbox
from tests.support.mock import patch


@pytest.fixture
def default_kwargs():
    return {
        "minion_id": "minion1",
        "pillar": None,
        "api_url": "http://netbox.example.com",
        "api_token": "yeic5oocizei7owuichoesh8ooqu6oob3uWiey9a",
        "api_query_result_limit": 65535,
    }


@pytest.fixture
def headers():
    return {"Authorization": "Token quin1Di5MoRooChaiph3Aenaxais5EeY1gie6eev"}


@pytest.fixture
def device_results():
    return {
        "dict": {
            "count": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 511,
                    "url": "https://netbox.example.com/api/dcim/devices/511/",
                    "name": "minion1",
                    "display_name": "minion1",
                    "device_type": {
                        "id": 4,
                        "url": "https://netbox.example.com/api/dcim/device-types/4/",
                        "manufacturer": {
                            "id": 1,
                            "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                            "name": "Cisco",
                            "slug": "cisco",
                        },
                        "model": "ISR2901",
                        "slug": "isr2901",
                        "display_name": "Cisco ISR2901",
                    },
                    "device_role": {
                        "id": 45,
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                        "name": "Network",
                        "slug": "network",
                    },
                    "node_type": "device",
                    "tenant": None,
                    "platform": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                        "name": "Cisco IOS",
                        "slug": "ios",
                    },
                    "serial": "",
                    "asset_tag": None,
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "rack": None,
                    "position": None,
                    "face": None,
                    "parent_device": None,
                    "status": {"value": "active", "label": "Active"},
                    "primary_ip": {
                        "id": 1146,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                        "family": 4,
                        "address": "192.0.2.1/24",
                    },
                    "primary_ip4": {
                        "id": 1146,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                        "family": 4,
                        "address": "192.0.2.1/24",
                    },
                    "primary_ip6": None,
                    "cluster": None,
                    "virtual_chassis": None,
                    "vc_position": None,
                    "vc_priority": None,
                    "comments": "",
                    "local_context_data": None,
                    "tags": [],
                    "custom_fields": {},
                    "config_context": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:04.171105Z",
                }
            ],
        }
    }


@pytest.fixture
def multiple_device_results():
    return {
        "dict": {
            "count": 2,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 511,
                    "url": "https://netbox.example.com/api/dcim/devices/511/",
                    "name": "minion1",
                    "display_name": "minion1",
                    "device_type": {
                        "id": 4,
                        "url": "https://netbox.example.com/api/dcim/device-types/4/",
                        "manufacturer": {
                            "id": 1,
                            "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                            "name": "Cisco",
                            "slug": "cisco",
                        },
                        "model": "ISR2901",
                        "slug": "isr2901",
                        "display_name": "Cisco ISR2901",
                    },
                    "device_role": {
                        "id": 45,
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                        "name": "Network",
                        "slug": "network",
                    },
                    "node_type": "device",
                    "tenant": None,
                    "platform": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                        "name": "Cisco IOS",
                        "slug": "ios",
                    },
                    "serial": "",
                    "asset_tag": None,
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "rack": None,
                    "position": None,
                    "face": None,
                    "parent_device": None,
                    "status": {"value": "active", "label": "Active"},
                    "primary_ip": {
                        "id": 1146,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                        "family": 4,
                        "address": "192.0.2.1/24",
                    },
                    "primary_ip4": {
                        "id": 1146,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                        "family": 4,
                        "address": "192.0.2.1/24",
                    },
                    "primary_ip6": None,
                    "cluster": None,
                    "virtual_chassis": None,
                    "vc_position": None,
                    "vc_priority": None,
                    "comments": "",
                    "local_context_data": None,
                    "tags": [],
                    "custom_fields": {},
                    "config_context": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:04.171105Z",
                },
                {
                    "id": 512,
                    "url": "https://netbox.example.com/api/dcim/devices/512/",
                    "name": "minion2",
                    "display_name": "minion2",
                    "device_type": {
                        "id": 4,
                        "url": "https://netbox.example.com/api/dcim/device-types/4/",
                        "manufacturer": {
                            "id": 1,
                            "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                            "name": "Cisco",
                            "slug": "cisco",
                        },
                        "model": "ISR2901",
                        "slug": "isr2901",
                        "display_name": "Cisco ISR2901",
                    },
                    "device_role": {
                        "id": 45,
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                        "name": "Network",
                        "slug": "network",
                    },
                    "node_type": "device",
                    "tenant": None,
                    "platform": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                        "name": "Cisco IOS",
                        "slug": "ios",
                    },
                    "serial": "",
                    "asset_tag": None,
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "rack": None,
                    "position": None,
                    "face": None,
                    "parent_device": None,
                    "status": {"value": "active", "label": "Active"},
                    "primary_ip": {
                        "id": 1150,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                        "family": 4,
                        "address": "192.0.2.3/24",
                    },
                    "primary_ip4": {
                        "id": 1150,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                        "family": 4,
                        "address": "192.0.2.3/24",
                    },
                    "primary_ip6": None,
                    "cluster": None,
                    "virtual_chassis": None,
                    "vc_position": None,
                    "vc_priority": None,
                    "comments": "",
                    "local_context_data": None,
                    "tags": [],
                    "custom_fields": {},
                    "config_context": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:04.171105Z",
                },
            ],
        }
    }


@pytest.fixture
def secondary_device_result():
    return {
        "dict": {
            "id": 512,
            "url": "https://netbox.example.com/api/dcim/devices/512/",
            "name": "minion2",
            "display_name": "minion2",
            "device_type": {
                "id": 4,
                "url": "https://netbox.example.com/api/dcim/device-types/4/",
                "manufacturer": {
                    "id": 1,
                    "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR2901",
                "slug": "isr2901",
                "display_name": "Cisco ISR2901",
            },
            "device_role": {
                "id": 45,
                "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                "name": "Network",
                "slug": "network",
            },
            "node_type": "device",
            "tenant": None,
            "platform": {
                "id": 1,
                "url": "https://netbox.example.com/api/dcim/platforms/1/",
                "name": "Cisco IOS",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 18,
                "url": "https://netbox.example.com/api/dcim/sites/18/",
                "name": "Site 1",
                "slug": "site1",
            },
            "rack": None,
            "position": None,
            "face": None,
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "primary_ip": {
                "id": 1150,
                "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                "family": 4,
                "address": "192.0.2.3/24",
            },
            "primary_ip4": {
                "id": 1150,
                "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                "family": 4,
                "address": "192.0.2.3/24",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-02-19",
            "last_updated": "2021-02-19T06:12:04.171105Z",
        }
    }


@pytest.fixture
def virtual_machine_results():
    return {
        "dict": {
            "count": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 222,
                    "url": "https://netbox.example.com/api/virtualization/virtual-machines/222/",
                    "name": "minion1",
                    "status": {"value": "active", "label": "Active"},
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "cluster": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/virtualization/clusters/1/",
                        "name": "Cluster",
                    },
                    "role": {
                        "id": 45,
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                        "name": "Network",
                        "slug": "network",
                    },
                    "node_type": "virtual-machine",
                    "tenant": None,
                    "platform": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                        "name": "Cisco IOS",
                        "slug": "ios",
                    },
                    "primary_ip": {
                        "id": 1148,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1148/",
                        "family": 4,
                        "address": "192.0.2.2/24",
                    },
                    "primary_ip4": {
                        "id": 1148,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1148/",
                        "family": 4,
                        "address": "192.0.2.2/24",
                    },
                    "primary_ip6": None,
                    "vcpus": 1,
                    "memory": 1024,
                    "disk": 30,
                    "comments": "",
                    "local_context_data": None,
                    "tags": [],
                    "custom_fields": {},
                    "config_context": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:05.799541Z",
                }
            ],
        }
    }


@pytest.fixture
def multiple_virtual_machine_results():
    return {
        "dict": {
            "count": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 222,
                    "url": "https://netbox.example.com/api/virtualization/virtual-machines/222/",
                    "name": "minion1",
                    "status": {"value": "active", "label": "Active"},
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "cluster": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/virtualization/clusters/1/",
                        "name": "Cluster",
                    },
                    "role": {
                        "id": 45,
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                        "name": "Network",
                        "slug": "network",
                    },
                    "node_type": "virtual-machine",
                    "tenant": None,
                    "platform": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                        "name": "Cisco IOS",
                        "slug": "ios",
                    },
                    "primary_ip": {
                        "id": 1148,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1148/",
                        "family": 4,
                        "address": "192.0.2.2/24",
                    },
                    "primary_ip4": {
                        "id": 1148,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1148/",
                        "family": 4,
                        "address": "192.0.2.2/24",
                    },
                    "primary_ip6": None,
                    "vcpus": 1,
                    "memory": 1024,
                    "disk": 30,
                    "comments": "",
                    "local_context_data": None,
                    "tags": [],
                    "custom_fields": {},
                    "config_context": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:05.799541Z",
                },
                {
                    "id": 223,
                    "url": "https://netbox.example.com/api/virtualization/virtual-machines/223/",
                    "name": "minion1",
                    "status": {"value": "active", "label": "Active"},
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "cluster": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/virtualization/clusters/1/",
                        "name": "Cluster",
                    },
                    "role": {
                        "id": 45,
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                        "name": "Network",
                        "slug": "network",
                    },
                    "node_type": "virtual-machine",
                    "tenant": None,
                    "platform": {
                        "id": 1,
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                        "name": "Cisco IOS",
                        "slug": "ios",
                    },
                    "primary_ip": {
                        "id": 1152,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1152/",
                        "family": 4,
                        "address": "192.0.2.4/24",
                    },
                    "primary_ip4": {
                        "id": 1152,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1152/",
                        "family": 4,
                        "address": "192.0.2.4/24",
                    },
                    "primary_ip6": None,
                    "vcpus": 1,
                    "memory": 1024,
                    "disk": 30,
                    "comments": "",
                    "local_context_data": None,
                    "tags": [],
                    "custom_fields": {},
                    "config_context": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:05.799541Z",
                },
            ],
        }
    }


@pytest.fixture
def no_results():
    return {"dict": {"count": 0, "next": None, "previous": None, "results": []}}


@pytest.fixture
def http_error():
    return {"error": "HTTP 404: Not Found", "status": 404}


@pytest.fixture
def device_interface_results():
    return {
        "dict": {
            "count": 2,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 8158,
                    "url": "https://netbox.example.com/api/dcim/interfaces/8158/",
                    "device": {
                        "id": 511,
                        "url": "https://netbox.example.com/api/dcim/devices/511/",
                        "name": "minion1",
                        "display_name": "minion1",
                    },
                    "name": "GigabitEthernet0/0",
                    "label": "",
                    "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
                    "enabled": True,
                    "lag": None,
                    "mtu": None,
                    "mac_address": None,
                    "mgmt_only": False,
                    "description": "",
                    "mode": None,
                    "untagged_vlan": None,
                    "tagged_vlans": [],
                    "cable": None,
                    "cable_peer": None,
                    "cable_peer_type": None,
                    "connected_endpoints": [
                        {
                            "id": 170,
                            "url": "https://demo.netbox.dev/api/dcim/interfaces/512/",
                            "display": "GigabitEthernet1/0/1",
                            "device": {
                                "id": 512,
                                "url": "https://demo.netbox.dev/api/dcim/devices/512/",
                                "display": "minion2",
                                "name": "minion2",
                            },
                            "name": "GigabitEthernet1/0/1",
                            "cable": 35,
                            "_occupied": True,
                        }
                    ],
                    "connected_endpoints_type": "dcim.interface",
                    "connected_endpoints_reachable": True,
                    "tags": [],
                    "count_ipaddresses": 1,
                },
                {
                    "id": 8159,
                    "url": "https://netbox.example.com/api/dcim/interfaces/8159/",
                    "device": {
                        "id": 511,
                        "url": "https://netbox.example.com/api/dcim/devices/511/",
                        "name": "minion1",
                        "display_name": "minion1",
                    },
                    "name": "GigabitEthernet0/1",
                    "label": "",
                    "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
                    "enabled": True,
                    "lag": None,
                    "mtu": None,
                    "mac_address": None,
                    "mgmt_only": False,
                    "description": "",
                    "mode": None,
                    "untagged_vlan": None,
                    "tagged_vlans": [],
                    "cable": None,
                    "cable_peer": None,
                    "cable_peer_type": None,
                    "connected_endpoints": None,
                    "connected_endpoints_type": None,
                    "connected_endpoints_reachable": None,
                    "tags": [],
                    "count_ipaddresses": 1,
                },
            ],
        }
    }


@pytest.fixture
def device_interfaces_list():
    return [
        {
            "id": 8158,
            "url": "https://netbox.example.com/api/dcim/interfaces/8158/",
            "name": "GigabitEthernet0/0",
            "label": "",
            "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
            "enabled": True,
            "lag": None,
            "mtu": None,
            "mac_address": None,
            "mgmt_only": False,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "cable": None,
            "cable_peer": None,
            "cable_peer_type": None,
            "connected_endpoints": [
                {
                    "_occupied": True,
                    "cable": 35,
                    "device": {
                        "display": "minion2",
                        "id": 512,
                        "name": "minion2",
                        "url": "https://demo.netbox.dev/api/dcim/devices/512/",
                    },
                    "display": "GigabitEthernet1/0/1",
                    "id": 170,
                    "name": "GigabitEthernet1/0/1",
                    "url": "https://demo.netbox.dev/api/dcim/interfaces/512/",
                }
            ],
            "connected_endpoints_reachable": True,
            "connected_endpoints_type": "dcim.interface",
            "tags": [],
            "count_ipaddresses": 1,
        },
        {
            "id": 8159,
            "url": "https://netbox.example.com/api/dcim/interfaces/8159/",
            "name": "GigabitEthernet0/1",
            "label": "",
            "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
            "enabled": True,
            "lag": None,
            "mtu": None,
            "mac_address": None,
            "mgmt_only": False,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "cable": None,
            "cable_peer": None,
            "cable_peer_type": None,
            "connected_endpoints": None,
            "connected_endpoints_type": None,
            "connected_endpoints_reachable": None,
            "tags": [],
            "count_ipaddresses": 1,
        },
    ]


@pytest.fixture
def virtual_machine_interface_results():
    return {
        "dict": {
            "count": 2,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 668,
                    "url": "https://netbox.example.com/api/virtualization/interfaces/668/",
                    "virtual_machine": {
                        "id": 222,
                        "url": "https://netbox.example.com/api/virtualization/virtual-machines/222/",
                        "name": "minion1",
                    },
                    "name": "GigabitEthernet0/0",
                    "enabled": True,
                    "mtu": None,
                    "mac_address": None,
                    "description": "",
                    "mode": None,
                    "untagged_vlan": None,
                    "tagged_vlans": [],
                    "tags": [],
                },
                {
                    "id": 669,
                    "url": "https://netbox.example.com/api/virtualization/interfaces/669/",
                    "virtual_machine": {
                        "id": 222,
                        "url": "https://netbox.example.com/api/virtualization/virtual-machines/222/",
                        "name": "minion1",
                    },
                    "name": "GigabitEthernet0/1",
                    "enabled": True,
                    "mtu": None,
                    "mac_address": None,
                    "description": "",
                    "mode": None,
                    "untagged_vlan": None,
                    "tagged_vlans": [],
                    "tags": [],
                },
            ],
        }
    }


@pytest.fixture
def virtual_machine_interfaces_list():
    return [
        {
            "id": 668,
            "url": "https://netbox.example.com/api/virtualization/interfaces/668/",
            "name": "GigabitEthernet0/0",
            "enabled": True,
            "mtu": None,
            "mac_address": None,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "tags": [],
        },
        {
            "id": 669,
            "url": "https://netbox.example.com/api/virtualization/interfaces/669/",
            "name": "GigabitEthernet0/1",
            "enabled": True,
            "mtu": None,
            "mac_address": None,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "tags": [],
        },
    ]


@pytest.fixture
def device_ip_results():
    return {
        "dict": {
            "count": 2,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 1146,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "192.0.2.1/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "assigned_object_type": "dcim.interface",
                    "assigned_object_id": 8158,
                    "assigned_object": {
                        "id": 8158,
                        "url": "https://netbox.example.com/api/dcim/interfaces/8158/",
                        "device": {
                            "id": 511,
                            "url": "https://netbox.example.com/api/dcim/devices/511/",
                            "name": "minion1",
                            "display_name": "minion1",
                        },
                        "name": "GigabitEthernet0/0",
                        "cable": None,
                    },
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:04.153386Z",
                },
                {
                    "id": 1147,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1147/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "198.51.100.1/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "assigned_object_type": "dcim.interface",
                    "assigned_object_id": 8159,
                    "assigned_object": {
                        "id": 8159,
                        "url": "https://netbox.example.com/api/dcim/interfaces/8159/",
                        "device": {
                            "id": 511,
                            "url": "https://netbox.example.com/api/dcim/devices/511/",
                            "name": "minion1",
                            "display_name": "minion1",
                        },
                        "name": "GigabitEthernet0/1",
                        "cable": None,
                    },
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:40.508154Z",
                },
            ],
        }
    }


@pytest.fixture
def virtual_machine_ip_results():
    return {
        "dict": {
            "count": 2,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 1148,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1148/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "192.0.2.2/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "assigned_object_type": "virtualization.vminterface",
                    "assigned_object_id": 668,
                    "assigned_object": {
                        "id": 668,
                        "url": "https://netbox.example.com/api/virtualization/interfaces/668/",
                        "virtual_machine": {
                            "id": 222,
                            "url": "https://netbox.example.com/api/virtualization/virtual-machines/222/",
                            "name": "minion1",
                        },
                        "name": "GigabitEthernet0/0",
                    },
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:05.784281Z",
                },
                {
                    "id": 1149,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1149/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "198.51.100.2/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "assigned_object_type": "virtualization.vminterface",
                    "assigned_object_id": 669,
                    "assigned_object": {
                        "id": 669,
                        "url": "https://netbox.example.com/api/virtualization/interfaces/669/",
                        "virtual_machine": {
                            "id": 222,
                            "url": "https://netbox.example.com/api/virtualization/virtual-machines/222/",
                            "name": "minion1",
                        },
                        "name": "GigabitEthernet0/1",
                    },
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:29.607428Z",
                },
            ],
        }
    }


@pytest.fixture
def device_interfaces_ip_list():
    return [
        {
            "id": 8158,
            "ip_addresses": [
                {
                    "id": 1146,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "192.0.2.1/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:04.153386Z",
                },
            ],
            "url": "https://netbox.example.com/api/dcim/interfaces/8158/",
            "name": "GigabitEthernet0/0",
            "label": "",
            "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
            "enabled": True,
            "lag": None,
            "mtu": None,
            "mac_address": None,
            "mgmt_only": False,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "cable": None,
            "cable_peer": None,
            "cable_peer_type": None,
            "connected_endpoints": [
                {
                    "id": 170,
                    "url": "https://demo.netbox.dev/api/dcim/interfaces/512/",
                    "display": "GigabitEthernet1/0/1",
                    "device": {
                        "id": 512,
                        "url": "https://demo.netbox.dev/api/dcim/devices/512/",
                        "display": "minion2",
                        "name": "minion2",
                    },
                    "name": "GigabitEthernet1/0/1",
                    "cable": 35,
                    "_occupied": True,
                }
            ],
            "connected_endpoints_type": "dcim.interface",
            "connected_endpoints_reachable": True,
            "tags": [],
            "count_ipaddresses": 1,
        },
        {
            "id": 8159,
            "ip_addresses": [
                {
                    "id": 1147,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1147/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "198.51.100.1/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:12:40.508154Z",
                },
            ],
            "url": "https://netbox.example.com/api/dcim/interfaces/8159/",
            "name": "GigabitEthernet0/1",
            "label": "",
            "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
            "enabled": True,
            "lag": None,
            "mtu": None,
            "mac_address": None,
            "mgmt_only": False,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "cable": None,
            "cable_peer": None,
            "cable_peer_type": None,
            "connected_endpoints": None,
            "connected_endpoints_type": None,
            "connected_endpoints_reachable": None,
            "tags": [],
            "count_ipaddresses": 1,
        },
    ]


@pytest.fixture
def virtual_machine_interfaces_ip_list():
    return [
        {
            "id": 668,
            "ip_addresses": [
                {
                    "id": 1148,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1148/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "192.0.2.2/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:05.784281Z",
                },
            ],
            "url": "https://netbox.example.com/api/virtualization/interfaces/668/",
            "name": "GigabitEthernet0/0",
            "enabled": True,
            "mtu": None,
            "mac_address": None,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "tags": [],
        },
        {
            "id": 669,
            "ip_addresses": [
                {
                    "id": 1149,
                    "url": "https://netbox.example.com/api/ipam/ip-addresses/1149/",
                    "family": {"value": 4, "label": "IPv4"},
                    "address": "198.51.100.2/24",
                    "vrf": None,
                    "tenant": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "nat_inside": None,
                    "nat_outside": None,
                    "dns_name": "",
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-19",
                    "last_updated": "2021-02-19T06:23:29.607428Z",
                },
            ],
            "url": "https://netbox.example.com/api/virtualization/interfaces/669/",
            "name": "GigabitEthernet0/1",
            "enabled": True,
            "mtu": None,
            "mac_address": None,
            "description": "",
            "mode": None,
            "untagged_vlan": None,
            "tagged_vlans": [],
            "tags": [],
        },
    ]


@pytest.fixture
def site_results():
    return {
        "dict": {
            "id": 18,
            "url": "https://netbox.example.com/api/dcim/sites/18/",
            "name": "Site 1",
            "slug": "site1",
            "status": {"value": "active", "label": "Active"},
            "region": None,
            "tenant": None,
            "facility": "",
            "asn": None,
            "time_zone": None,
            "description": "",
            "physical_address": "",
            "shipping_address": "",
            "latitude": None,
            "longitude": None,
            "contact_name": "",
            "contact_phone": "",
            "contact_email": "",
            "comments": "",
            "tags": [],
            "custom_fields": {},
            "created": "2021-02-25",
            "last_updated": "2021-02-25T14:21:07.898957Z",
            "circuit_count": 0,
            "device_count": 1,
            "prefix_count": 2,
            "rack_count": 0,
            "virtualmachine_count": 1,
            "vlan_count": 0,
        }
    }


@pytest.fixture
def site_prefixes_results():
    return {
        "dict": {
            "count": 2,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": 284,
                    "url": "https://netbox.example.com/api/ipam/prefixes/284/",
                    "family": {"value": 4, "label": "IPv4"},
                    "prefix": "192.0.2.0/24",
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "vrf": None,
                    "tenant": None,
                    "vlan": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "is_pool": False,
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-25",
                    "last_updated": "2021-02-25T15:08:27.136305Z",
                },
                {
                    "id": 285,
                    "url": "https://netbox.example.com/api/ipam/prefixes/285/",
                    "family": {"value": 4, "label": "IPv4"},
                    "prefix": "198.51.100.0/24",
                    "site": {
                        "id": 18,
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                        "name": "Site 1",
                        "slug": "site1",
                    },
                    "vrf": None,
                    "tenant": None,
                    "vlan": None,
                    "status": {"value": "active", "label": "Active"},
                    "role": None,
                    "is_pool": False,
                    "description": "",
                    "tags": [],
                    "custom_fields": {},
                    "created": "2021-02-25",
                    "last_updated": "2021-02-25T15:08:59.880440Z",
                },
            ],
        }
    }


@pytest.fixture
def site_prefixes():
    return [
        {
            "id": 284,
            "url": "https://netbox.example.com/api/ipam/prefixes/284/",
            "family": {"value": 4, "label": "IPv4"},
            "prefix": "192.0.2.0/24",
            "vrf": None,
            "tenant": None,
            "vlan": None,
            "status": {"value": "active", "label": "Active"},
            "role": None,
            "is_pool": False,
            "description": "",
            "tags": [],
            "custom_fields": {},
            "created": "2021-02-25",
            "last_updated": "2021-02-25T15:08:27.136305Z",
        },
        {
            "id": 285,
            "url": "https://netbox.example.com/api/ipam/prefixes/285/",
            "family": {"value": 4, "label": "IPv4"},
            "prefix": "198.51.100.0/24",
            "vrf": None,
            "tenant": None,
            "vlan": None,
            "status": {"value": "active", "label": "Active"},
            "role": None,
            "is_pool": False,
            "description": "",
            "tags": [],
            "custom_fields": {},
            "created": "2021-02-25",
            "last_updated": "2021-02-25T15:08:59.880440Z",
        },
    ]


@pytest.fixture
def proxy_details_results():
    return {
        "dict": {
            "id": 1,
            "url": "https://netbox.example.com/api/dcim/platforms/1/",
            "name": "Cisco IOS",
            "slug": "ios",
            "manufacturer": {
                "id": 1,
                "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                "name": "Cisco",
                "slug": "cisco",
            },
            "napalm_driver": "ios",
            "napalm_args": None,
            "description": "",
            "device_count": 152,
            "virtualmachine_count": 1,
        }
    }


@pytest.fixture
def proxy_details():
    return {
        "host": "192.0.2.1",
        "driver": "ios",
        "proxytype": "napalm",
    }


@pytest.fixture
def pillar_results():
    return {
        "netbox": {
            "id": 511,
            "url": "https://netbox.example.com/api/dcim/devices/511/",
            "name": "minion1",
            "node_type": "device",
            "display_name": "minion1",
            "device_type": {
                "id": 4,
                "url": "https://netbox.example.com/api/dcim/device-types/4/",
                "manufacturer": {
                    "id": 1,
                    "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR2901",
                "slug": "isr2901",
                "display_name": "Cisco ISR2901",
            },
            "device_role": {
                "id": 45,
                "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                "name": "Network",
                "slug": "network",
            },
            "interfaces": [
                {
                    "id": 8158,
                    "ip_addresses": [
                        {
                            "id": 1146,
                            "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                            "family": {"value": 4, "label": "IPv4"},
                            "address": "192.0.2.1/24",
                            "vrf": None,
                            "tenant": None,
                            "status": {"value": "active", "label": "Active"},
                            "role": None,
                            "nat_inside": None,
                            "nat_outside": None,
                            "dns_name": "",
                            "description": "",
                            "tags": [],
                            "custom_fields": {},
                            "created": "2021-02-19",
                            "last_updated": "2021-02-19T06:12:04.153386Z",
                        },
                    ],
                    "url": "https://netbox.example.com/api/dcim/interfaces/8158/",
                    "name": "GigabitEthernet0/0",
                    "label": "",
                    "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
                    "enabled": True,
                    "lag": None,
                    "mtu": None,
                    "mac_address": None,
                    "mgmt_only": False,
                    "description": "",
                    "mode": None,
                    "untagged_vlan": None,
                    "tagged_vlans": [],
                    "cable": None,
                    "cable_peer": None,
                    "cable_peer_type": None,
                    "connected_endpoints": [
                        {
                            "id": 170,
                            "url": "https://demo.netbox.dev/api/dcim/interfaces/512/",
                            "display": "GigabitEthernet1/0/1",
                            "device": {
                                "id": 512,
                                "url": "https://demo.netbox.dev/api/dcim/devices/512/",
                                "display": "minion2",
                                "name": "minion2",
                            },
                            "name": "GigabitEthernet1/0/1",
                            "cable": 35,
                            "_occupied": True,
                        }
                    ],
                    "connected_endpoints_type": "dcim.interface",
                    "connected_endpoints_reachable": True,
                    "tags": [],
                    "count_ipaddresses": 1,
                },
                {
                    "id": 8159,
                    "ip_addresses": [
                        {
                            "id": 1147,
                            "url": "https://netbox.example.com/api/ipam/ip-addresses/1147/",
                            "family": {"value": 4, "label": "IPv4"},
                            "address": "198.51.100.1/24",
                            "vrf": None,
                            "tenant": None,
                            "status": {"value": "active", "label": "Active"},
                            "role": None,
                            "nat_inside": None,
                            "nat_outside": None,
                            "dns_name": "",
                            "description": "",
                            "tags": [],
                            "custom_fields": {},
                            "created": "2021-02-19",
                            "last_updated": "2021-02-19T06:12:40.508154Z",
                        },
                    ],
                    "url": "https://netbox.example.com/api/dcim/interfaces/8159/",
                    "name": "GigabitEthernet0/1",
                    "label": "",
                    "type": {"value": "1000base-t", "label": "1000BASE-T (1GE)"},
                    "enabled": True,
                    "lag": None,
                    "mtu": None,
                    "mac_address": None,
                    "mgmt_only": False,
                    "description": "",
                    "mode": None,
                    "untagged_vlan": None,
                    "tagged_vlans": [],
                    "cable": None,
                    "cable_peer": None,
                    "cable_peer_type": None,
                    "connected_endpoints": None,
                    "connected_endpoints_type": None,
                    "connected_endpoints_reachable": None,
                    "tags": [],
                    "count_ipaddresses": 1,
                },
            ],
            "tenant": None,
            "platform": {
                "id": 1,
                "url": "https://netbox.example.com/api/dcim/platforms/1/",
                "name": "Cisco IOS",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 18,
                "url": "https://netbox.example.com/api/dcim/sites/18/",
                "name": "Site 1",
                "slug": "site1",
                "status": {"value": "active", "label": "Active"},
                "region": None,
                "tenant": None,
                "facility": "",
                "asn": None,
                "time_zone": None,
                "description": "",
                "physical_address": "",
                "shipping_address": "",
                "latitude": None,
                "longitude": None,
                "contact_name": "",
                "contact_phone": "",
                "contact_email": "",
                "comments": "",
                "tags": [],
                "custom_fields": {},
                "created": "2021-02-25",
                "last_updated": "2021-02-25T14:21:07.898957Z",
                "circuit_count": 0,
                "device_count": 1,
                "prefix_count": 2,
                "rack_count": 0,
                "virtualmachine_count": 1,
                "vlan_count": 0,
                "prefixes": [
                    {
                        "id": 284,
                        "url": "https://netbox.example.com/api/ipam/prefixes/284/",
                        "family": {"value": 4, "label": "IPv4"},
                        "prefix": "192.0.2.0/24",
                        "vrf": None,
                        "tenant": None,
                        "vlan": None,
                        "status": {"value": "active", "label": "Active"},
                        "role": None,
                        "is_pool": False,
                        "description": "",
                        "tags": [],
                        "custom_fields": {},
                        "created": "2021-02-25",
                        "last_updated": "2021-02-25T15:08:27.136305Z",
                    },
                    {
                        "id": 285,
                        "url": "https://netbox.example.com/api/ipam/prefixes/285/",
                        "family": {"value": 4, "label": "IPv4"},
                        "prefix": "198.51.100.0/24",
                        "vrf": None,
                        "tenant": None,
                        "vlan": None,
                        "status": {"value": "active", "label": "Active"},
                        "role": None,
                        "is_pool": False,
                        "description": "",
                        "tags": [],
                        "custom_fields": {},
                        "created": "2021-02-25",
                        "last_updated": "2021-02-25T15:08:59.880440Z",
                    },
                ],
            },
            "rack": None,
            "position": None,
            "face": None,
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "primary_ip": {
                "id": 1146,
                "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                "family": 4,
                "address": "192.0.2.1/24",
            },
            "primary_ip4": {
                "id": 1146,
                "url": "https://netbox.example.com/api/ipam/ip-addresses/1146/",
                "family": 4,
                "address": "192.0.2.1/24",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "connected_devices": {
                512: {
                    "asset_tag": None,
                    "cluster": None,
                    "comments": "",
                    "config_context": {},
                    "created": "2021-02-19",
                    "custom_fields": {},
                    "device_role": {
                        "id": 45,
                        "name": "Network",
                        "slug": "network",
                        "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                    },
                    "device_type": {
                        "display_name": "Cisco " "ISR2901",
                        "id": 4,
                        "manufacturer": {
                            "id": 1,
                            "name": "Cisco",
                            "slug": "cisco",
                            "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                        },
                        "model": "ISR2901",
                        "slug": "isr2901",
                        "url": "https://netbox.example.com/api/dcim/device-types/4/",
                    },
                    "display_name": "minion2",
                    "face": None,
                    "id": 512,
                    "last_updated": "2021-02-19T06:12:04.171105Z",
                    "local_context_data": None,
                    "name": "minion2",
                    "node_type": "device",
                    "parent_device": None,
                    "platform": {
                        "id": 1,
                        "name": "Cisco IOS",
                        "slug": "ios",
                        "url": "https://netbox.example.com/api/dcim/platforms/1/",
                    },
                    "position": None,
                    "primary_ip": {
                        "address": "192.0.2.3/24",
                        "family": 4,
                        "id": 1150,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                    },
                    "primary_ip4": {
                        "address": "192.0.2.3/24",
                        "family": 4,
                        "id": 1150,
                        "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                    },
                    "primary_ip6": None,
                    "rack": None,
                    "serial": "",
                    "site": {
                        "id": 18,
                        "name": "Site 1",
                        "slug": "site1",
                        "url": "https://netbox.example.com/api/dcim/sites/18/",
                    },
                    "status": {"label": "Active", "value": "active"},
                    "tags": [],
                    "tenant": None,
                    "url": "https://netbox.example.com/api/dcim/devices/512/",
                    "vc_position": None,
                    "vc_priority": None,
                    "virtual_chassis": None,
                }
            },
            "created": "2021-02-19",
            "last_updated": "2021-02-19T06:12:04.171105Z",
        },
        "proxy": {"host": "192.0.2.1", "driver": "ios", "proxytype": "napalm"},
    }


@pytest.fixture
def connected_devices_results():
    return {
        512: {
            "id": 512,
            "url": "https://netbox.example.com/api/dcim/devices/512/",
            "name": "minion2",
            "display_name": "minion2",
            "device_type": {
                "id": 4,
                "url": "https://netbox.example.com/api/dcim/device-types/4/",
                "manufacturer": {
                    "id": 1,
                    "url": "https://netbox.example.com/api/dcim/manufacturers/1/",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR2901",
                "slug": "isr2901",
                "display_name": "Cisco ISR2901",
            },
            "device_role": {
                "id": 45,
                "url": "https://netbox.example.com/api/dcim/device-roles/45/",
                "name": "Network",
                "slug": "network",
            },
            "node_type": "device",
            "tenant": None,
            "platform": {
                "id": 1,
                "url": "https://netbox.example.com/api/dcim/platforms/1/",
                "name": "Cisco IOS",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 18,
                "url": "https://netbox.example.com/api/dcim/sites/18/",
                "name": "Site 1",
                "slug": "site1",
            },
            "rack": None,
            "position": None,
            "face": None,
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "primary_ip": {
                "id": 1150,
                "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                "family": 4,
                "address": "192.0.2.3/24",
            },
            "primary_ip4": {
                "id": 1150,
                "url": "https://netbox.example.com/api/ipam/ip-addresses/1150/",
                "family": 4,
                "address": "192.0.2.3/24",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-02-19",
            "last_updated": "2021-02-19T06:12:04.171105Z",
        }
    }


def test_when_minion_id_is_star_then_result_should_be_empty_dict(default_kwargs):
    expected_result = {}
    default_kwargs["minion_id"] = "*"

    actual_result = netbox.ext_pillar(**default_kwargs)

    assert actual_result == expected_result


def test_when_api_url_is_not_http_or_https_then_error_message_should_be_logged(
    default_kwargs,
):
    default_kwargs["api_url"] = "ftp://netbox.example.com"

    with patch("salt.pillar.netbox.log.error", autospec=True) as fake_error:
        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'Provided URL for api_url "%s" is malformed or is not an http/https URL',
            "ftp://netbox.example.com",
        )


def test_when_neither_devices_or_virtual_machines_requested_then_error_message_should_be_logged(
    default_kwargs,
):
    default_kwargs["devices"] = default_kwargs["virtual_machines"] = False

    with patch("salt.pillar.netbox.log.error", autospec=True) as fake_error:
        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            "At least one of devices or virtual_machines must be True"
        )


def test_when_interface_ips_requested_but_not_interfaces_then_error_message_should_be_logged(
    default_kwargs,
):
    default_kwargs["interfaces"] = False
    default_kwargs["interface_ips"] = True

    with patch("salt.pillar.netbox.log.error", autospec=True) as fake_error:
        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            "The value for interfaces must be True if interface_ips is True"
        )


def test_when_api_query_result_limit_set_but_not_a_positive_integer_then_error_message_should_be_logged(
    default_kwargs,
):
    default_kwargs["api_query_result_limit"] = -1

    with patch("salt.pillar.netbox.log.error", autospec=True) as fake_error:
        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            "The value for api_query_result_limit must be a postive integer if set"
        )


def test_when_api_token_not_set_then_error_message_should_be_logged(
    default_kwargs,
):

    default_kwargs["api_token"] = ""

    with patch("salt.pillar.netbox.log.error", autospec=True) as fake_error:
        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with("The value for api_token is not set")


def test_when_we_retrieve_a_single_device_then_return_list(
    default_kwargs, headers, device_results
):

    expected_result = device_results["dict"]["results"]

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = device_results

        actual_result = netbox._get_devices(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_a_device_and_get_http_error_then_return_empty_list(
    default_kwargs, headers, http_error
):
    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_devices(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_a_single_virtual_machine_then_return_list(
    default_kwargs, headers, virtual_machine_results
):

    expected_result = virtual_machine_results["dict"]["results"]

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = virtual_machine_results

        actual_result = netbox._get_virtual_machines(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_a_virtual_machine_and_get_http_error_then_return_empty_dict(
    default_kwargs, headers, http_error
):

    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_virtual_machines(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_device_interfaces_then_return_dict(
    default_kwargs, headers, device_interface_results, device_interfaces_list
):

    expected_result = device_interfaces_list

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = device_interface_results

        actual_result = netbox._get_interfaces(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            511,
            "device",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_device_interfaces_and_get_http_error_then_return_empty_list(
    default_kwargs, headers, http_error
):

    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_interfaces(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            511,
            "device",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_virtual_machine_interfaces_then_return_list(
    default_kwargs,
    headers,
    virtual_machine_interface_results,
    virtual_machine_interfaces_list,
):

    expected_result = virtual_machine_interfaces_list

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = virtual_machine_interface_results

        actual_result = netbox._get_interfaces(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            222,
            "virtual-machine",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_virtual_machine_interfaces_and_get_http_error_then_return_empty_list(
    default_kwargs, headers, http_error
):

    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_interfaces(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            222,
            "virtual-machine",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_device_interface_ips_then_return_list(
    default_kwargs, headers, device_ip_results
):

    expected_result = device_ip_results["dict"]["results"]

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = device_ip_results

        actual_result = netbox._get_interface_ips(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            511,
            "device",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_connected_endpoints(
    default_kwargs,
    headers,
    connected_devices_results,
    device_interfaces_list,
    secondary_device_result,
):

    expected_result = connected_devices_results

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = secondary_device_result

        actual_result = netbox._get_connected_devices(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            device_interfaces_list,
            headers,
        )

        assert actual_result == expected_result


def test_when_we_retrieve_device_interface_ips_and_get_http_error_then_return_empty_list(
    default_kwargs, headers, http_error
):

    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_interface_ips(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            511,
            "device",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_virtual_machine_interface_ips_then_return_list(
    default_kwargs, headers, virtual_machine_ip_results
):

    expected_result = virtual_machine_ip_results["dict"]["results"]

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = virtual_machine_ip_results

        actual_result = netbox._get_interface_ips(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            222,
            "virtual-machine",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_virtual_machine_interface_ips_and_get_http_error_then_return_empty_list(
    default_kwargs, headers, http_error
):

    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_interface_ips(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            222,
            "virtual-machine",
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_associate_ips_to_interfaces_then_return_list(
    default_kwargs, device_interfaces_list, device_ip_results, device_interfaces_ip_list
):

    expected_result = device_interfaces_ip_list

    interfaces_list = device_interfaces_list
    interface_ips_list = device_ip_results["dict"]["results"]

    actual_result = netbox._associate_ips_to_interfaces(
        interfaces_list, interface_ips_list
    )

    assert actual_result == expected_result


def test_associate_empty_ip_list_to_interfaces_then_return_list(
    default_kwargs, device_interfaces_list, device_ip_results
):

    expected_result = device_interfaces_list

    interfaces_list = device_interfaces_list
    interface_ips_list = []

    actual_result = netbox._associate_ips_to_interfaces(
        interfaces_list, interface_ips_list
    )

    assert actual_result == expected_result


def test_when_we_retrieve_site_details_then_return_dict(
    default_kwargs, headers, site_results
):

    expected_result = site_results["dict"]

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = site_results

        actual_result = netbox._get_site_details(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            "Site 1",
            18,
            headers,
        )

        assert actual_result == expected_result


def test_when_we_retrieve_site_details_and_get_http_error_then_return_empty_dict(
    default_kwargs, headers, http_error
):

    expected_result = {}

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_site_details(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            "Site 1",
            18,
            headers,
        )

        assert actual_result == expected_result


def test_when_we_retrieve_site_prefixes_then_return_list(
    default_kwargs, headers, site_prefixes_results, site_prefixes
):

    expected_result = site_prefixes

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = site_prefixes_results

        actual_result = netbox._get_site_prefixes(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            "Site 1",
            18,
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_site_prefixes_and_get_http_error_then_return_empty_list(
    default_kwargs, headers, http_error
):

    expected_result = []

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_site_prefixes(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            "Site 1",
            18,
            headers,
            default_kwargs["api_query_result_limit"],
        )

        assert actual_result == expected_result


def test_when_we_retrieve_proxy_details_then_return_dict(
    default_kwargs, headers, proxy_details_results, proxy_details
):

    expected_result = proxy_details

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = proxy_details_results

        actual_result = netbox._get_proxy_details(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            "192.0.2.1/24",
            1,
            headers,
        )

        assert actual_result == expected_result


def test_when_we_retrieve_proxy_details_and_get_http_error_then_dont_return(
    default_kwargs, headers, http_error
):

    expected_result = None

    with patch("salt.utils.http.query", autospec=True) as query:
        query.return_value = http_error

        actual_result = netbox._get_proxy_details(
            default_kwargs["api_url"],
            default_kwargs["minion_id"],
            "192.0.2.1/24",
            1,
            headers,
        )

        assert actual_result == expected_result


def test_when_we_retrieve_multiple_devices_then_error_message_should_be_logged(
    default_kwargs, multiple_device_results
):

    with patch(
        "salt.pillar.netbox._get_devices", autospec=True
    ) as multiple_devices, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:

        multiple_devices.return_value = multiple_device_results["dict"]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'More than one node found for "%s"',
            "minion1",
        )


def test_when_we_retrieve_multiple_virtual_machines_then_error_message_should_be_logged(
    default_kwargs, multiple_virtual_machine_results
):
    default_kwargs["devices"] = False
    default_kwargs["virtual_machines"] = True

    with patch(
        "salt.pillar.netbox._get_virtual_machines", autospec=True
    ) as multiple_virtual_machines, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:

        multiple_virtual_machines.return_value = multiple_virtual_machine_results[
            "dict"
        ]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'More than one node found for "%s"',
            "minion1",
        )


def test_when_we_retrieve_a_device_and_a_virtual_machine_then_error_message_should_be_logged(
    default_kwargs, device_results, virtual_machine_results
):
    default_kwargs["virtual_machines"] = True

    with patch("salt.pillar.netbox._get_devices", autospec=True) as device, patch(
        "salt.pillar.netbox._get_virtual_machines", autospec=True
    ) as virtual_machine, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:

        device.return_value = device_results["dict"]["results"]
        virtual_machine.return_value = virtual_machine_results["dict"]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'More than one node found for "%s"',
            "minion1",
        )


def test_when_we_retrieve_no_devices_then_error_message_should_be_logged(
    default_kwargs, no_results
):

    with patch("salt.pillar.netbox._get_devices", autospec=True) as devices, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:

        devices.return_value = no_results["dict"]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'Unable to pull NetBox data for "%s"',
            "minion1",
        )


def test_when_we_retrieve_no_virtual_machines_then_error_message_should_be_logged(
    default_kwargs, no_results
):
    default_kwargs["devices"] = False
    default_kwargs["virtual_machines"] = True

    with patch(
        "salt.pillar.netbox._get_virtual_machines", autospec=True
    ) as virtual_machines, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:
        virtual_machines.return_value = no_results["dict"]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'Unable to pull NetBox data for "%s"',
            "minion1",
        )


def test_when_we_retrieve_everything_successfully_then_return_dict(
    default_kwargs,
    device_results,
    no_results,
    device_interfaces_list,
    device_ip_results,
    site_results,
    site_prefixes,
    proxy_details,
    pillar_results,
    connected_devices_results,
):

    expected_result = pillar_results

    default_kwargs["virtual_machines"] = False
    default_kwargs["interfaces"] = True
    default_kwargs["interface_ips"] = True
    default_kwargs["site_details"] = True
    default_kwargs["site_prefixes"] = True
    default_kwargs["proxy_return"] = True
    default_kwargs["connected_devices"] = True

    with patch("salt.pillar.netbox._get_devices", autospec=True) as get_devices, patch(
        "salt.pillar.netbox._get_virtual_machines", autospec=True
    ) as get_virtual_machines, patch(
        "salt.pillar.netbox._get_interfaces", autospec=True
    ) as get_interfaces, patch(
        "salt.pillar.netbox._get_interface_ips", autospec=True
    ) as get_interface_ips, patch(
        "salt.pillar.netbox._get_site_details", autospec=True
    ) as get_site_details, patch(
        "salt.pillar.netbox._get_site_prefixes", autospec=True
    ) as get_site_prefixes, patch(
        "salt.pillar.netbox._get_proxy_details", autospec=True
    ) as get_proxy_details, patch(
        "salt.pillar.netbox._get_connected_devices", autospec=True
    ) as get_connected_decvices:

        get_devices.return_value = device_results["dict"]["results"]
        get_virtual_machines.return_value = no_results["dict"]["results"]
        get_interfaces.return_value = device_interfaces_list
        get_interface_ips.return_value = device_ip_results["dict"]["results"]
        get_site_details.return_value = site_results["dict"]
        get_site_prefixes.return_value = site_prefixes
        get_proxy_details.return_value = proxy_details
        get_connected_decvices.return_value = connected_devices_results

        actual_result = netbox.ext_pillar(**default_kwargs)

        assert actual_result == expected_result


def test_when_we_set_proxy_return_but_get_no_value_for_platform_then_error_message_should_be_logged(
    default_kwargs, headers, device_results
):

    default_kwargs["site_details"] = False
    default_kwargs["site_prefixes"] = False
    default_kwargs["proxy_return"] = True
    device_results["dict"]["results"][0]["platform"] = None

    with patch("salt.pillar.netbox._get_devices", autospec=True) as devices, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:

        devices.return_value = device_results["dict"]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'You have set "proxy_return" to "True" but you have not set the platform in NetBox for "%s"',
            "minion1",
        )


def test_when_we_set_proxy_return_but_get_no_value_for_primary_ip_then_error_message_should_be_logged(
    default_kwargs, headers, device_results
):

    default_kwargs["site_details"] = False
    default_kwargs["site_prefixes"] = False
    default_kwargs["proxy_return"] = True
    device_results["dict"]["results"][0]["primary_ip"] = None

    with patch("salt.pillar.netbox._get_devices", autospec=True) as devices, patch(
        "salt.pillar.netbox.log.error", autospec=True
    ) as fake_error:

        devices.return_value = device_results["dict"]["results"]

        netbox.ext_pillar(**default_kwargs)

        fake_error.assert_called_with(
            'You have set "proxy_return" to "True" but you have not set the primary IPv4 or IPv6 address in NetBox for "%s"',
            "minion1",
        )



"""
Unit tests for salt.transport.base.
"""

import pytest

import salt.transport.base

pytestmark = [
    pytest.mark.core_test,
]


def test_unclosed_warning():

    transport = salt.transport.base.Transport()
    assert transport._closing is False
    assert transport._connect_called is False
    transport.connect()
    assert transport._connect_called is True
    with pytest.warns(salt.transport.base.TransportWarning):
        del transport



"""
    salt._logging.impl
    ~~~~~~~~~~~~~~~~~~

    Salt's logging implementation classes/functionality
"""

import atexit
import logging
import multiprocessing
import os
import pathlib
import re
import socket
import sys
import traceback
import types
import urllib.parse

# Let's define these custom logging levels before importing the salt._logging.mixins
# since they will be used there
PROFILE = logging.PROFILE = 15
TRACE = logging.TRACE = 5
GARBAGE = logging.GARBAGE = 1
QUIET = logging.QUIET = 1000

import salt.defaults.exitcodes  # isort:skip  pylint: disable=unused-import
from salt._logging.handlers import DeferredStreamHandler  # isort:skip
from salt._logging.handlers import RotatingFileHandler  # isort:skip
from salt._logging.handlers import StreamHandler  # isort:skip
from salt._logging.handlers import SysLogHandler  # isort:skip
from salt._logging.handlers import WatchedFileHandler  # isort:skip
from salt._logging.mixins import LoggingMixinMeta  # isort:skip
from salt.exceptions import LoggingRuntimeError  # isort:skip
from salt.utils.ctx import RequestContext  # isort:skip
from salt.utils.immutabletypes import freeze, ImmutableDict  # isort:skip
from salt.utils.textformat import TextFormat  # isort:skip

LOG_LEVELS = {
    "all": logging.NOTSET,
    "debug": logging.DEBUG,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "garbage": GARBAGE,
    "info": logging.INFO,
    "profile": PROFILE,
    "quiet": QUIET,
    "trace": TRACE,
    "warning": logging.WARNING,
}

LOG_VALUES_TO_LEVELS = {v: k for (k, v) in LOG_LEVELS.items()}

LOG_COLORS = {
    "levels": {
        "QUIET": TextFormat("reset"),
        "CRITICAL": TextFormat("bold", "red"),
        "ERROR": TextFormat("bold", "red"),
        "WARNING": TextFormat("bold", "yellow"),
        "INFO": TextFormat("bold", "green"),
        "PROFILE": TextFormat("bold", "cyan"),
        "DEBUG": TextFormat("bold", "cyan"),
        "TRACE": TextFormat("bold", "magenta"),
        "GARBAGE": TextFormat("bold", "blue"),
        "NOTSET": TextFormat("reset"),
        "SUBDEBUG": TextFormat(
            "bold", "cyan"
        ),  # used by multiprocessing.log_to_stderr()
        "SUBWARNING": TextFormat(
            "bold", "yellow"
        ),  # used by multiprocessing.log_to_stderr()
    },
    "msgs": {
        "QUIET": TextFormat("reset"),
        "CRITICAL": TextFormat("bold", "red"),
        "ERROR": TextFormat("red"),
        "WARNING": TextFormat("yellow"),
        "INFO": TextFormat("green"),
        "PROFILE": TextFormat("bold", "cyan"),
        "DEBUG": TextFormat("cyan"),
        "TRACE": TextFormat("magenta"),
        "GARBAGE": TextFormat("blue"),
        "NOTSET": TextFormat("reset"),
        "SUBDEBUG": TextFormat(
            "bold", "cyan"
        ),  # used by multiprocessing.log_to_stderr()
        "SUBWARNING": TextFormat(
            "bold", "yellow"
        ),  # used by multiprocessing.log_to_stderr()
    },
    "name": TextFormat("bold", "green"),
    "process": TextFormat("bold", "blue"),
}

# Make a list of log level names sorted by log level
SORTED_LEVEL_NAMES = [l[0] for l in sorted(LOG_LEVELS.items(), key=lambda x: x[1])]

MODNAME_PATTERN = re.compile(r"(?P<name>%%\(name\)(?:\-(?P<digits>[\d]+))?s)")

# Default logging formatting options
DFLT_LOG_FMT_JID = "[JID: %(jid)s]"
DFLT_LOG_DATEFMT = "%H:%M:%S"
DFLT_LOG_DATEFMT_LOGFILE = "%Y-%m-%d %H:%M:%S"
DFLT_LOG_FMT_CONSOLE = "[%(levelname)-8s] %(message)s"
DFLT_LOG_FMT_LOGFILE = "%(asctime)s,%(msecs)03d [%(name)-17s:%(lineno)-4d][%(levelname)-8s][%(process)d] %(message)s"


class SaltLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        logging.LogRecord.__init__(self, *args, **kwargs)
        self.bracketname = f"[{str(self.name):<17}]"
        self.bracketlevel = f"[{str(self.levelname):<8}]"
        self.bracketprocess = f"[{str(self.process):>5}]"


class SaltColorLogRecord(SaltLogRecord):
    def __init__(self, *args, **kwargs):
        SaltLogRecord.__init__(self, *args, **kwargs)

        reset = TextFormat("reset")
        clevel = LOG_COLORS["levels"].get(self.levelname, reset)
        cmsg = LOG_COLORS["msgs"].get(self.levelname, reset)

        self.colorname = "{}[{:<17}]{}".format(
            LOG_COLORS["name"], str(self.name), reset
        )
        self.colorlevel = f"{clevel}[{str(self.levelname):<8}]{reset}"
        self.colorprocess = "{}[{:>5}]{}".format(
            LOG_COLORS["process"], str(self.process), reset
        )
        self.colormsg = f"{cmsg}{self.getMessage()}{reset}"


def get_log_record_factory():
    """
    Get the logging  log record factory
    """
    try:
        return get_log_record_factory.__factory__
    except AttributeError:
        return


def set_log_record_factory(factory):
    """
    Set the logging  log record factory
    """
    get_log_record_factory.__factory__ = factory
    logging.setLogRecordFactory(factory)


set_log_record_factory(SaltLogRecord)


# Store an instance of the current logging logger class
LOGGING_LOGGER_CLASS = logging.getLoggerClass()


class SaltLoggingClass(LOGGING_LOGGER_CLASS, metaclass=LoggingMixinMeta):

    ONCECACHE = set()

    def __new__(cls, *args):
        """
        We override `__new__` in our logging logger class in order to provide
        some additional features like expand the module name padding if length
        is being used, and also some Unicode fixes.

        This code overhead will only be executed when the class is
        instantiated, i.e.:

            logging.getLogger(__name__)

        """
        instance = super().__new__(cls)  # pylint: disable=no-value-for-parameter

        try:
            max_logger_length = len(
                max(list(logging.Logger.manager.loggerDict), key=len)
            )
            if max_logger_length > 80:
                # Make sure the logger name on the formatted log record is not longer than 80 chars
                # Messages which need more that 80 chars will use them, but not ALL log messages
                max_logger_length = 80
            for handler in logging.root.handlers:
                if handler is get_temp_handler():
                    continue

                formatter = handler.formatter
                if not formatter:
                    continue

                if not handler.lock:
                    handler.createLock()
                handler.acquire()

                fmt = formatter._fmt.replace("%", "%%")

                match = MODNAME_PATTERN.search(fmt)
                if not match:
                    # Not matched. Release handler and return.
                    handler.release()
                    return instance

                if "digits" not in match.groupdict():
                    # No digits group. Release handler and return.
                    handler.release()
                    return instance

                digits = match.group("digits")
                if not digits or not (digits and digits.isdigit()):
                    # No valid digits. Release handler and return.
                    handler.release()
                    return instance

                if int(digits) < max_logger_length:
                    # Formatter digits value is lower than current max, update.
                    fmt = fmt.replace(match.group("name"), "%%(name)-%ds")
                    formatter = logging.Formatter(
                        fmt % max_logger_length, datefmt=formatter.datefmt
                    )
                    handler.setFormatter(formatter)
                handler.release()
        except ValueError:
            # There are no registered loggers yet
            pass
        return instance

    def _log(
        self,
        level,
        msg,
        args,
        exc_info=None,
        extra=None,  # pylint: disable=arguments-differ
        stack_info=False,
        stacklevel=1,
        exc_info_on_loglevel=None,
        once=False,
    ):
        if extra is None:
            extra = {}

        if once:
            if str(args) in self.ONCECACHE:
                return
            self.ONCECACHE.add(str(args))

        # pylint: disable=no-member
        current_jid = RequestContext.current.get("data", {}).get("jid", None)
        log_fmt_jid = RequestContext.current.get("opts", {}).get("log_fmt_jid", None)
        # pylint: enable=no-member

        if current_jid is not None:
            extra["jid"] = current_jid

        if log_fmt_jid is not None:
            extra["log_fmt_jid"] = log_fmt_jid

        # If both exc_info and exc_info_on_loglevel are both passed, let's fail
        if exc_info and exc_info_on_loglevel:
            raise LoggingRuntimeError(
                "Only one of 'exc_info' and 'exc_info_on_loglevel' is permitted"
            )
        if exc_info_on_loglevel is not None:
            if isinstance(exc_info_on_loglevel, str):
                exc_info_on_loglevel = LOG_LEVELS.get(
                    exc_info_on_loglevel, logging.ERROR
                )
            elif not isinstance(exc_info_on_loglevel, int):
                raise RuntimeError(
                    "The value of 'exc_info_on_loglevel' needs to be a "
                    "logging level or a logging level name, not '{}'".format(
                        exc_info_on_loglevel
                    )
                )
        # XXX: extra is never None
        if extra is None:
            extra = {"exc_info_on_loglevel": exc_info_on_loglevel}
        else:
            extra["exc_info_on_loglevel"] = exc_info_on_loglevel

        # this is required for log lines to work as expected because we are
        # adding a stackframe with this function
        stacklevel = stacklevel + 1

        try:
            LOGGING_LOGGER_CLASS._log(
                self,
                level,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
                stacklevel=stacklevel,
            )
        except TypeError:
            # stacklevel was introduced in Py 3.8
            # must be running on old OS with Python 3.6 or 3.7
            LOGGING_LOGGER_CLASS._log(
                self,
                level,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )

    def makeRecord(
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func=None,
        extra=None,
        sinfo=None,
    ):
        # Let's remove exc_info_on_loglevel from extra
        exc_info_on_loglevel = extra.pop("exc_info_on_loglevel")

        jid = extra.pop("jid", "")
        if jid:
            log_fmt_jid = extra.pop("log_fmt_jid")
            jid = log_fmt_jid % {"jid": jid}

        if not extra:
            # If nothing else is in extra, make it None
            extra = None

        # Let's try to make every logging message unicode
        try:
            salt_system_encoding = __salt_system_encoding__
            if salt_system_encoding == "ascii":
                # Encoding detection most likely failed, let's use the utf-8
                # value which we defaulted before __salt_system_encoding__ was
                # implemented
                salt_system_encoding = "utf-8"
        except NameError:
            salt_system_encoding = "utf-8"

        if isinstance(msg, bytes):
            try:
                _msg = msg.decode(salt_system_encoding, "replace")
            except UnicodeDecodeError:
                _msg = msg.decode(salt_system_encoding, "ignore")
        else:
            _msg = msg

        _args = []
        for item in args:
            if isinstance(item, bytes):
                try:
                    _args.append(item.decode(salt_system_encoding, "replace"))
                except UnicodeDecodeError:
                    _args.append(item.decode(salt_system_encoding, "ignore"))
            else:
                _args.append(item)
        _args = tuple(_args)

        logrecord = LOGGING_LOGGER_CLASS.makeRecord(
            self, name, level, fn, lno, _msg, _args, exc_info, func, sinfo
        )

        if exc_info_on_loglevel is not None:
            # Let's add some custom attributes to the LogRecord class in order
            # to include the exc_info on a per handler basis. This will allow
            # showing tracebacks on logfiles but not on console if the logfile
            # handler is enabled for the log level "exc_info_on_loglevel" and
            # console handler is not.
            logrecord.exc_info_on_loglevel_instance = sys.exc_info()
            logrecord.exc_info_on_loglevel_formatted = None

        logrecord.exc_info_on_loglevel = exc_info_on_loglevel
        logrecord.jid = jid
        return logrecord


def get_logging_level_from_string(level):
    """
    Return an integer matching a logging level.
    Return logging.ERROR when not matching.
    Return logging.WARNING when the passed level is None
    """
    if level is None:
        return logging.WARNING

    if isinstance(level, int):
        # Level is already an integer, return it
        return level
    try:
        return LOG_LEVELS[level.lower()]
    except KeyError:
        if level:
            log.warning(
                "Could not translate the logging level string '%s' "
                "into an actual logging level integer. Returning "
                "'logging.ERROR'.",
                level,
            )
        # Couldn't translate the passed string into a logging level.
        return logging.ERROR


def get_logging_options_dict():
    """
    Return the logging options dictionary
    """
    try:
        return set_logging_options_dict.__options_dict__
    except AttributeError:
        return


def set_logging_options_dict(opts):
    """
    Create a logging related options dictionary based off of the loaded salt config
    """
    try:
        if isinstance(set_logging_options_dict.__options_dict__, ImmutableDict):
            raise RuntimeError(
                "The logging options have been frozen. They can no longer be changed."
            )
    except AttributeError:
        pass
    set_logging_options_dict.__options_dict__ = opts
    set_lowest_log_level_by_opts(opts)


def freeze_logging_options_dict():
    """
    Turn the logging options dictionary into an immutable dictionary
    """
    set_logging_options_dict.__options_dict__ = freeze(
        set_logging_options_dict.__options_dict__
    )


def is_temp_handler_configured():
    """
    Is the temporary deferred stream handler configured
    """
    return get_temp_handler() is not None


def get_temp_handler():
    """
    Get the temporary deferred stream handler
    """
    try:
        return setup_temp_handler.__handler__
    except AttributeError:
        return


def setup_temp_handler(log_level=None):
    """
    Setup the temporary deferred stream handler
    """
    handler = get_temp_handler()
    if handler is not None:
        log_level = get_logging_level_from_string(log_level)
        if handler.level != log_level:
            handler.setLevel(log_level)
        return

    if log_level is None:
        log_level = logging.WARNING

    log_level = get_logging_level_from_string(log_level)

    handler = None
    for handler in logging.root.handlers:

        if not hasattr(handler, "stream"):
            # Not a stream handler, continue
            continue

        if handler.stream is sys.stderr:
            # There's already a logging handler outputting to sys.stderr
            break
    else:
        handler = DeferredStreamHandler(sys.stderr)
        atexit.register(handler.flush)
    handler.setLevel(log_level)

    # Set the default temporary console formatter config
    formatter = logging.Formatter(DFLT_LOG_FMT_CONSOLE, datefmt=DFLT_LOG_DATEFMT)
    handler.setFormatter(formatter)
    logging.root.addHandler(handler)

    setup_temp_handler.__handler__ = handler


def shutdown_temp_handler():
    """
    Shutdown the temporary deferred stream handler
    """
    temp_handler = get_temp_handler()
    if temp_handler is not None:
        for handler in logging.root.handlers[:]:
            if handler is temp_handler:
                logging.root.handlers.remove(handler)
                handler.sync_with_handlers(logging.root.handlers)
                handler.close()
                break
        # Redefine the handler to None so it can be garbage collected
        setup_temp_handler.__handler__ = None


# Override the python's logging logger class as soon as this module is imported
if logging.getLoggerClass() is not SaltLoggingClass:

    # Import pip._internal which itself will install it's own custom logging handler
    # we want to override that handler with ours
    try:
        import pip._internal.utils._log as pip_log_module  # pylint: disable=no-name-in-module,import-error
    except ImportError:
        pip_log_module = None

    logging.setLoggerClass(SaltLoggingClass)
    logging.addLevelName(QUIET, "QUIET")
    logging.addLevelName(PROFILE, "PROFILE")
    logging.addLevelName(TRACE, "TRACE")
    logging.addLevelName(GARBAGE, "GARBAGE")
    if pip_log_module is not None:
        # Let's make newer versions of pip work by patching SaltLoggingClass to
        # add a verbose method which is what pip expects
        SaltLoggingClass.verbose = SaltLoggingClass.debug

    if not logging.root.handlers:
        # No configuration to the logging system has been done so far.
        # Set the root logger at the lowest level possible
        logging.root.setLevel(GARBAGE)

        # Add a permanent null handler so that we never get messages like:
        #   No handlers could be found for logger 'foo'
        setup_temp_handler()
        logging.root.addHandler(get_temp_handler())


# Now that we defined the default logging logger class, we can instantiate our logger
# DO NOT MOVE THIS
log = logging.getLogger(__name__)


def get_console_handler():
    """
    Get the console stream handler
    """
    try:
        return setup_console_handler.__handler__
    except AttributeError:
        return


def is_console_handler_configured():
    """
    Is the console stream handler configured
    """
    return get_console_handler() is not None


def shutdown_console_handler():
    """
    Shutdown the console stream handler
    """
    console_handler = get_console_handler()
    if console_handler is not None:
        logging.root.removeHandler(console_handler)
        console_handler.close()
        setup_console_handler.__handler__ = None
        atexit.unregister(shutdown_console_handler)


def setup_console_handler(log_level=None, log_format=None, date_format=None):
    """
    Setup the console stream handler
    """
    if is_console_handler_configured():
        log.warning("Console logging already configured")
        return

    atexit.register(shutdown_console_handler)

    log.trace(
        "Setting up console logging: %s",
        dict(log_level=log_level, log_format=log_format, date_format=date_format),
    )

    if log_level is None:
        log_level = logging.WARNING

    log_level = get_logging_level_from_string(log_level)

    set_log_record_factory(SaltColorLogRecord)

    handler = None
    for handler in logging.root.handlers:
        if handler is get_temp_handler():
            continue

        if not hasattr(handler, "stream"):
            # Not a stream handler, continue
            continue

        if handler.stream is sys.stderr:
            # There's already a logging handler outputting to sys.stderr
            break
    else:
        handler = StreamHandler(sys.stderr)
    handler.setLevel(log_level)

    # Set the default console formatter config
    if not log_format:
        log_format = DFLT_LOG_FMT_CONSOLE
    if not date_format:
        date_format = DFLT_LOG_DATEFMT

    formatter = logging.Formatter(log_format, datefmt=date_format)

    handler.setFormatter(formatter)
    logging.root.addHandler(handler)

    setup_console_handler.__handler__ = handler


def get_logfile_handler():
    """
    Get the log file handler
    """
    try:
        return setup_logfile_handler.__handler__
    except AttributeError:
        return


def is_logfile_handler_configured():
    """
    Is the log file handler configured
    """
    return get_logfile_handler() is not None


def shutdown_logfile_handler():
    """
    Shutdown the log file handler
    """
    logfile_handler = get_logfile_handler()
    if logfile_handler is not None:
        logging.root.removeHandler(logfile_handler)
        logfile_handler.close()
        setup_logfile_handler.__handler__ = None
        atexit.unregister(shutdown_logfile_handler)


def setup_logfile_handler(
    log_path,
    log_level=None,
    log_format=None,
    date_format=None,
    max_bytes=0,
    backup_count=0,
    user=None,
):
    """
    Setup the log file handler

    Since version 0.10.6 we support logging to syslog, some examples:

        tcp://localhost:514/LOG_USER
        tcp://localhost/LOG_DAEMON
        udp://localhost:5145/LOG_KERN
        udp://localhost
        file:///dev/log
        file:///dev/log/LOG_SYSLOG
        file:///dev/log/LOG_DAEMON

    The above examples are self explanatory, but:
        <file|udp|tcp>://<host|socketpath>:<port-if-required>/<log-facility>

    Thinking on doing remote logging you might also be thinking that
    you could point Salt's logging to the remote syslog. **Please Don't!**
    An issue has been reported when doing this over TCP where the logged lines
    get concatenated. See #3061.

    The preferred way to do remote logging is setup a local syslog, point
    Salt's logging to the local syslog(unix socket is much faster) and then
    have the local syslog forward the log messages to the remote syslog.
    """
    if is_logfile_handler_configured():
        log.warning("Logfile logging already configured")
        return

    atexit.register(shutdown_logfile_handler)
    log.trace(
        "Setting up log file logging: %s",
        dict(
            log_path=log_path,
            log_level=log_level,
            log_format=log_format,
            date_format=date_format,
            max_bytes=max_bytes,
            backup_count=backup_count,
            user=user,
        ),
    )

    if log_path is None:
        log.warning("log_path setting is set to `None`. Nothing else to do")
        return

    if log_level is None:
        log_level = logging.WARNING

    log_level = get_logging_level_from_string(log_level)

    parsed_log_path = urllib.parse.urlparse(log_path)

    if parsed_log_path.scheme in ("tcp", "udp", "file"):
        syslog_opts = {
            "facility": SysLogHandler.LOG_USER,
            "socktype": socket.SOCK_DGRAM,
        }

        if parsed_log_path.scheme == "file" and parsed_log_path.path:
            path = pathlib.Path(parsed_log_path.path)
            facility_name = path.stem.upper()
            try:
                if not facility_name.startswith("LOG_"):
                    # The user is not specifying a syslog facility
                    facility_name = "LOG_USER"  # Syslog default
                    syslog_opts["address"] = str(path.resolve())
                else:
                    # The user has set a syslog facility, let's update the path to
                    # the logging socket
                    syslog_opts["address"] = str(path.resolve().parent)
            except OSError as exc:
                raise LoggingRuntimeError(
                    f"Failed to setup the Syslog logging handler: {exc}"
                ) from exc
        elif parsed_log_path.path:
            # In case of udp or tcp with a facility specified
            path = pathlib.Path(parsed_log_path.path)
            facility_name = path.stem.upper()
            if not facility_name.startswith("LOG_"):
                # Logging facilities start with LOG_ if this is not the case
                # fail right now!
                raise LoggingRuntimeError(
                    f"The syslog facility '{facility_name}' is not known"
                )
        else:
            # This is the case of udp or tcp without a facility specified
            facility_name = "LOG_USER"  # Syslog default

        facility = getattr(SysLogHandler, facility_name, None)
        if facility is None:
            # This python syslog version does not know about the user provided
            # facility name
            raise LoggingRuntimeError(
                f"The syslog facility '{facility_name}' is not known"
            )
        syslog_opts["facility"] = facility

        if parsed_log_path.scheme in ("tcp", "udp"):
            syslog_opts["address"] = (
                parsed_log_path.hostname,
                parsed_log_path.port or logging.handlers.SYSLOG_UDP_PORT,
            )
            if parsed_log_path.scheme == "tcp":
                syslog_opts["socktype"] = socket.SOCK_STREAM

        elif parsed_log_path.scheme == "file":
            syslog_opts.pop("socktype", None)

        try:
            # Et voilá! Finally our syslog handler instance
            handler = SysLogHandler(**syslog_opts)
        except OSError as exc:
            raise LoggingRuntimeError(
                f"Failed to setup the Syslog logging handler: {exc}"
            ) from exc
    else:
        # make sure, the logging directory exists and attempt to create it if necessary
        if user is None:
            import salt.utils.user

            user = salt.utils.user.get_user()

        import salt.utils.files
        import salt.utils.verify

        # Logfile is not using Syslog, verify
        with salt.utils.files.set_umask(0o027):
            salt.utils.verify.verify_log_files([log_path], user)
        try:
            # Logfile logging is UTF-8 on purpose.
            # Since salt uses YAML and YAML uses either UTF-8 or UTF-16, if a
            # user is not using plain ASCII, their system should be ready to
            # handle UTF-8.
            if max_bytes > 0:
                handler = RotatingFileHandler(
                    log_path,
                    mode="a",
                    maxBytes=max_bytes,
                    backupCount=backup_count,
                    encoding="utf-8",
                    delay=0,
                )
            else:
                handler = WatchedFileHandler(
                    log_path, mode="a", encoding="utf-8", delay=0
                )
        except OSError:
            log.warning(
                "Failed to open log file, do you have permission to write to %s?",
                log_path,
            )
            # Do not proceed with any more configuration since it will fail, we
            # have the console logging already setup and the user should see
            # the error.
            return

    handler.setLevel(log_level)

    if not log_format:
        log_format = DFLT_LOG_FMT_LOGFILE
    if not date_format:
        date_format = DFLT_LOG_DATEFMT_LOGFILE

    formatter = logging.Formatter(log_format, datefmt=date_format)

    handler.setFormatter(formatter)
    logging.root.addHandler(handler)

    setup_logfile_handler.__handler__ = handler


def get_extended_logging_handlers():
    """
    Get the extended logging handlers
    """
    try:
        return setup_extended_logging.__handlers__
    except AttributeError:
        return


def is_extended_logging_configured():
    """
    Are the extended logging handlers configured
    """
    extended_logging_handlers = get_extended_logging_handlers()
    if extended_logging_handlers is None:
        return False
    return True


def shutdown_extended_logging():
    """
    Shutdown the extended logging handlers
    """
    extended_logging_handlers = get_extended_logging_handlers()
    if extended_logging_handlers:
        for handler in extended_logging_handlers:
            logging.root.removeHandler(handler)
            handler.close()
        atexit.unregister(shutdown_extended_logging)
    setup_extended_logging.__handlers__ = None


def setup_extended_logging(opts):
    """
    Setup the extended logging handlers, internal or external
    """
    if is_extended_logging_configured() is True:
        # Don't re-configure external loggers
        return

    # Explicit late import of Salt's loader
    import salt.loader

    # Be sure to always shutdown extened logging on process termination
    atexit.register(shutdown_extended_logging)

    # Let's keep a reference to the current logging handlers
    initial_handlers = logging.root.handlers[:]

    # Load any additional logging handlers
    providers = salt.loader.log_handlers(opts)

    # Let's keep track of the new logging handlers so we can sync the stored
    # log records with them
    additional_handlers = []

    for name, get_handlers_func in providers.items():
        log.info("Processing 'log_handlers.%s'", name)
        # Keep a reference to the logging handlers count before getting the
        # possible additional ones.
        initial_handlers_count = len(logging.root.handlers)

        handlers = get_handlers_func()
        if isinstance(handlers, types.GeneratorType):
            handlers = list(handlers)
        elif handlers is False or handlers == [False]:
            # A false return value means not configuring any logging handler on
            # purpose
            log.info(
                "The `log_handlers.%s.setup_handlers()` function returned "
                "`False` which means no logging handler was configured on "
                "purpose. Continuing...",
                name,
            )
            continue
        else:
            # Make sure we have an iterable
            handlers = [handlers]

        for handler in handlers:
            if not handler and len(logging.root.handlers) == initial_handlers_count:
                log.info(
                    "The `log_handlers.%s`, did not return any handlers and the "
                    "global handlers count did not increase. This could be a sign "
                    "that `log_handlers.%s` is not working as expected.",
                    name,
                    name,
                )
                continue

            log.debug("Adding the '%s' provided logging handler: '%s'", name, handler)
            additional_handlers.append(handler)
            logging.root.addHandler(handler)

    for handler in logging.root.handlers:
        if handler in initial_handlers:
            continue
        additional_handlers.append(handler)

    setup_extended_logging.__handlers__ = additional_handlers


def setup_log_granular_levels(log_granular_levels):
    """
    Get the extended logging handlers
    """
    for handler_name, handler_level in log_granular_levels.items():
        _logger = logging.getLogger(handler_name)
        _logger.setLevel(get_logging_level_from_string(handler_level))


def setup_logging():
    opts = get_logging_options_dict()
    if not opts:
        raise RuntimeError("The logging options have not been set yet.")
    if (
        opts.get("configure_console_logger", True)
        and not is_console_handler_configured()
    ):
        setup_console_handler(
            log_level=opts["log_level"],
            log_format=opts["log_fmt_console"],
            date_format=opts["log_datefmt"],
        )
    if opts.get("configure_file_logger", True) and not is_logfile_handler_configured():
        log_file_level = opts["log_level_logfile"] or opts["log_level"]
        if log_file_level != "quiet":
            log_file_key = opts.get("log_file_key") or "log_file"
            setup_logfile_handler(
                log_path=opts[log_file_key],
                log_level=log_file_level,
                log_format=opts["log_fmt_logfile"],
                date_format=opts["log_datefmt_logfile"],
                max_bytes=opts["log_rotate_max_bytes"],
                backup_count=opts["log_rotate_backup_count"],
                user=opts["user"],
            )
        else:
            setup_logfile_handler.__handler__ = logging.NullHandler()
    if (
        opts.get("configure_ext_handlers", True)
        and not is_extended_logging_configured()
    ):
        setup_extended_logging(opts)
    if opts.get("configure_granular_levels", True):
        setup_log_granular_levels(opts["log_granular_levels"])

    # Any logging that should be configured, is configured by now. Shutdown the temporary logging handler.
    shutdown_temp_handler()


def shutdown_logging():
    if is_temp_handler_configured():
        shutdown_temp_handler()
    if is_extended_logging_configured():
        shutdown_extended_logging()
    if is_logfile_handler_configured():
        shutdown_logfile_handler()
    if is_console_handler_configured():
        shutdown_console_handler()


def in_mainprocess():
    """
    Check to see if this is the main process
    """
    try:
        return in_mainprocess.__pid__ == os.getpid()
    except AttributeError:
        if multiprocessing.current_process().name == "MainProcess":
            in_mainprocess.__pid__ = os.getpid()
            return True
        return False


def get_lowest_log_level():
    """
    Get the lowest log level
    """
    try:
        return set_lowest_log_level.__log_level__
    except AttributeError:
        return


def set_lowest_log_level(log_level):
    """
    Set the lowest log level
    """
    set_lowest_log_level.__log_level__ = log_level
    # Additionally set the root logger to the same level.
    # Nothing below this level should be processed by python's logging machinery
    logging.root.setLevel(log_level)


def set_lowest_log_level_by_opts(opts):
    """
    Set the lowest log level by the passed config
    """
    log_levels = [
        get_logging_level_from_string(opts.get("log_level")),
        get_logging_level_from_string(opts.get("log_level_logfile")),
    ]
    for log_level in opts.get("log_granular_levels", {}).values():
        log_levels.append(get_logging_level_from_string(log_level))

    log_level = min(log_levels)
    set_lowest_log_level(log_level)


def __global_logging_exception_handler(
    exc_type,
    exc_value,
    exc_traceback,
    _logger=logging.getLogger(__name__),
    _stderr=sys.__stderr__,
    _format_exception=traceback.format_exception,
):
    """
    This function will log all un-handled python exceptions.
    """
    if exc_type.__name__ == "KeyboardInterrupt":
        # Do not log the exception or display the traceback on Keyboard Interrupt
        return

    # Log the exception
    msg = "An un-handled exception was caught by Salt's global exception handler:"
    try:
        msg = "{}\n{}: {}\n{}".format(
            msg,
            exc_type.__name__,
            exc_value,
            "".join(_format_exception(exc_type, exc_value, exc_traceback)).strip(),
        )
    except Exception:  # pylint: disable=broad-except
        msg = "{}\n{}: {}\n(UNABLE TO FORMAT TRACEBACK)".format(
            msg,
            exc_type.__name__,
            exc_value,
        )
    try:
        _logger.error(msg)
    except Exception:  # pylint: disable=broad-except
        # Python is shutting down and logging has been set to None already
        try:
            _stderr.write(msg + "\n")
        except Exception:  # pylint: disable=broad-except
            # We have also lost reference to sys.__stderr__ ?!
            print(msg)

    # Call the original sys.excepthook
    try:
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
    except Exception:  # pylint: disable=broad-except
        # Python is shutting down and sys has been set to None already
        pass


# Set our own exception handler as the one to use
sys.excepthook = __global_logging_exception_handler



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.win_autoruns as win_autoruns
from tests.support.mock import MagicMock, patch


@pytest.fixture
def define_key():
    return [
        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run /reg:64",
        "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
    ]


@pytest.fixture
def configure_loader_modules():
    return {win_autoruns: {}}


def test_list(define_key):
    """
    Test if it enables win_autoruns the service on the server
    """
    with patch("os.listdir", MagicMock(return_value=[])):
        ret = {
            define_key[0]: ["SALT"],
            define_key[1]: ["SALT"],
            define_key[2]: ["SALT"],
        }
        mock = MagicMock(return_value="Windows 7")
        with patch.dict(win_autoruns.__grains__, {"osfullname": mock}):
            mock = MagicMock(return_value="SALT")
            with patch.dict(win_autoruns.__salt__, {"cmd.run": mock}):
                assert win_autoruns.list_() == ret



import pytest

import salt.executors.splay as splay_exec


@pytest.fixture
def configure_loader_modules():
    return {splay_exec: {"__grains__": {"id": "foo"}}}


def test__get_hash():
    # We just want to make sure that this function does not result in an
    # error due to passing a unicode value to bytearray()
    assert splay_exec._get_hash()



"""
Management of timezones
=======================

The timezone can be managed for the system:

.. code-block:: yaml

    America/Denver:
      timezone.system

The system and the hardware clock are not necessarily set to the same time.
By default, the hardware clock is set to localtime, meaning it is set to the
same time as the system clock. If `utc` is set to True, then the hardware clock
will be set to UTC, and the system clock will be an offset of that.

.. code-block:: yaml

    America/Denver:
      timezone.system:
        - utc: True

.. _here: https://help.ubuntu.com/community/UbuntuTime#Multiple_Boot_Systems_Time_Conflicts

The Ubuntu community documentation contains an explanation of this setting, as
it applies to systems that dual-boot with Windows. This is explained in greater
detail here_.
"""

from salt.exceptions import CommandExecutionError, SaltInvocationError


def __virtual__():
    """
    Only load if the timezone module is available in __salt__
    """
    if "timezone.get_zone" in __salt__:
        return True
    return (False, "timezone module could not be loaded")


def system(name, utc=True):
    """
    Set the timezone for the system.

    name
        The name of the timezone to use (e.g.: America/Denver)

    utc
        Whether or not to set the hardware clock to UTC (default is True)
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    # Set up metadata
    do_utc = False
    do_zone = False

    try:
        compzone = __salt__["timezone.zone_compare"](name)
    except (SaltInvocationError, CommandExecutionError) as exc:
        ret["result"] = False
        ret["comment"] = (
            "Unable to compare desired timezone '{}' to system timezone: {}".format(
                name, exc
            )
        )
        return ret

    myutc = True
    messages = []
    if __salt__["timezone.get_hwclock"]() == "localtime":
        myutc = False

    # Check the time zone
    if compzone is True:
        ret["result"] = True
        messages.append(f"Timezone {name} already set")
    else:
        do_zone = True

    # If the user passed in utc, do a check
    if utc and utc != myutc:
        ret["result"] = None
        do_utc = True
    elif utc and utc == myutc:
        messages.append(f"UTC already set to {name}")

    if ret["result"] is True:
        ret["comment"] = ", ".join(messages)
        return ret

    if __opts__["test"]:
        messages = []
        if compzone is False:
            messages.append(f"Timezone {name} needs to be set")
        if utc and myutc != utc:
            messages.append(f"UTC needs to be set to {utc}")
        ret["comment"] = ", ".join(messages)
        return ret

    messages = []

    if do_zone:
        if __salt__["timezone.set_zone"](name):
            ret["changes"]["timezone"] = name
            messages.append(f"Set timezone {name}")
            ret["result"] = True
        else:
            messages.append("Failed to set timezone")
            ret["result"] = False

    if do_utc:
        clock = "localtime"
        if utc:
            clock = "UTC"
        if __salt__["timezone.set_hwclock"](clock):
            ret["changes"]["utc"] = utc
            messages.append(f"Set UTC to {utc}")
            ret["result"] = True
        else:
            messages.append(f"Failed to set UTC to {utc}")
            ret["result"] = False

    ret["comment"] = ", ".join(messages)
    return ret



"""
Cisco IOS configuration manipulation helpers

.. versionadded:: 2019.2.0

This module provides a collection of helper functions for Cisco IOS style
configuration manipulation. This module does not have external dependencies
and can be used from any Proxy or regular Minion.
"""

import difflib

# Import Salt modules
from collections import OrderedDict

import salt.utils.dictdiffer
import salt.utils.dictupdate
from salt.exceptions import SaltException

# ------------------------------------------------------------------------------
# module properties
# ------------------------------------------------------------------------------

__virtualname__ = "iosconfig"
__proxyenabled__ = ["*"]

# ------------------------------------------------------------------------------
# helper functions -- will not be exported
# ------------------------------------------------------------------------------


def _attach_data_to_path(obj, ele, data):
    if ele not in obj:
        obj[ele] = OrderedDict()
        obj[ele] = data
    else:
        obj[ele].update(data)


def _attach_data_to_path_tags(obj, path, data, list_=False):
    if "#list" not in obj:
        obj["#list"] = []
    path = [path]
    obj_tmp = obj
    first = True
    while True:
        obj_tmp["#text"] = " ".join(path)
        path_item = path.pop(0)
        if not path:
            break
        else:
            if path_item not in obj_tmp:
                obj_tmp[path_item] = OrderedDict()
            obj_tmp = obj_tmp[path_item]

            if first and list_:
                obj["#list"].append({path_item: obj_tmp})
                first = False
    if path_item in obj_tmp:
        obj_tmp[path_item].update(data)
    else:
        obj_tmp[path_item] = data
    obj_tmp[path_item]["#standalone"] = True


def _parse_text_config(config_lines, with_tags=False, current_indent=0, nested=False):
    struct_cfg = OrderedDict()
    while config_lines:
        line = config_lines.pop(0)
        if not line.strip() or line.lstrip().startswith("!"):
            # empty or comment
            continue
        current_line = line.lstrip()
        leading_spaces = len(line) - len(current_line)
        if leading_spaces > current_indent:
            current_block = _parse_text_config(
                config_lines,
                current_indent=leading_spaces,
                with_tags=with_tags,
                nested=True,
            )
            if with_tags:
                _attach_data_to_path_tags(
                    struct_cfg, current_line, current_block, nested
                )
            else:
                _attach_data_to_path(struct_cfg, current_line, current_block)
        elif leading_spaces < current_indent:
            config_lines.insert(0, line)
            break
        else:
            if not nested:
                current_block = _parse_text_config(
                    config_lines,
                    current_indent=leading_spaces,
                    with_tags=with_tags,
                    nested=True,
                )
                if with_tags:
                    _attach_data_to_path_tags(
                        struct_cfg, current_line, current_block, nested
                    )
                else:
                    _attach_data_to_path(struct_cfg, current_line, current_block)
            else:
                config_lines.insert(0, line)
                break
    return struct_cfg


def _get_diff_text(old, new):
    """
    Returns the diff of two text blobs.
    """
    diff = difflib.unified_diff(old.splitlines(1), new.splitlines(1))
    return "".join([x.replace("\r", "") for x in diff])


def _print_config_text(tree, indentation=0):
    """
    Return the config as text from a config tree.
    """
    config = ""
    for key, value in tree.items():
        config += "{indent}{line}\n".format(indent=" " * indentation, line=key)
        if value:
            config += _print_config_text(value, indentation=indentation + 1)
    return config


# ------------------------------------------------------------------------------
# callable functions
# ------------------------------------------------------------------------------


def tree(config=None, path=None, with_tags=False, saltenv="base"):
    """
    Transform Cisco IOS style configuration to structured Python dictionary.
    Depending on the value of the ``with_tags`` argument, this function may
    provide different views, valuable in different situations.

    config
        The configuration sent as text. This argument is ignored when ``path``
        is configured.

    path
        Absolute or remote path from where to load the configuration text. This
        argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    with_tags: ``False``
        Whether this function should return a detailed view, with tags.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.tree path=salt://path/to/my/config.txt
        salt '*' iosconfig.tree path=https://bit.ly/2mAdq7z
    """
    if path:
        config = __salt__["cp.get_file_str"](path, saltenv=saltenv)
        if config is False:
            raise SaltException(f"{path} is not available")
    config_lines = config.splitlines()
    return _parse_text_config(config_lines, with_tags=with_tags)


def clean(config=None, path=None, saltenv="base"):
    """
    Return a clean version of the config, without any special signs (such as
    ``!`` as an individual line) or empty lines, but just lines with significant
    value in the configuration of the network device.

    config
        The configuration sent as text. This argument is ignored when ``path``
        is configured.

    path
        Absolute or remote path from where to load the configuration text. This
        argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.clean path=salt://path/to/my/config.txt
        salt '*' iosconfig.clean path=https://bit.ly/2mAdq7z
    """
    config_tree = tree(config=config, path=path, saltenv=saltenv)
    return _print_config_text(config_tree)


def merge_tree(
    initial_config=None,
    initial_path=None,
    merge_config=None,
    merge_path=None,
    saltenv="base",
):
    """
    Return the merge tree of the ``initial_config`` with the ``merge_config``,
    as a Python dictionary.

    initial_config
        The initial configuration sent as text. This argument is ignored when
        ``initial_path`` is set.

    initial_path
        Absolute or remote path from where to load the initial configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    merge_config
        The config to be merged into the initial config, sent as text. This
        argument is ignored when ``merge_path`` is set.

    merge_path
        Absolute or remote path from where to load the merge configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``initial_path`` or ``merge_path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.merge_tree initial_path=salt://path/to/running.cfg merge_path=salt://path/to/merge.cfg
    """
    merge_tree = tree(config=merge_config, path=merge_path, saltenv=saltenv)
    initial_tree = tree(config=initial_config, path=initial_path, saltenv=saltenv)
    return salt.utils.dictupdate.merge(initial_tree, merge_tree)


def merge_text(
    initial_config=None,
    initial_path=None,
    merge_config=None,
    merge_path=None,
    saltenv="base",
):
    """
    Return the merge result of the ``initial_config`` with the ``merge_config``,
    as plain text.

    initial_config
        The initial configuration sent as text. This argument is ignored when
        ``initial_path`` is set.

    initial_path
        Absolute or remote path from where to load the initial configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    merge_config
        The config to be merged into the initial config, sent as text. This
        argument is ignored when ``merge_path`` is set.

    merge_path
        Absolute or remote path from where to load the merge configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``initial_path`` or ``merge_path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.merge_text initial_path=salt://path/to/running.cfg merge_path=salt://path/to/merge.cfg
    """
    candidate_tree = merge_tree(
        initial_config=initial_config,
        initial_path=initial_path,
        merge_config=merge_config,
        merge_path=merge_path,
        saltenv=saltenv,
    )
    return _print_config_text(candidate_tree)


def merge_diff(
    initial_config=None,
    initial_path=None,
    merge_config=None,
    merge_path=None,
    saltenv="base",
):
    """
    Return the merge diff, as text, after merging the merge config into the
    initial config.

    initial_config
        The initial configuration sent as text. This argument is ignored when
        ``initial_path`` is set.

    initial_path
        Absolute or remote path from where to load the initial configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    merge_config
        The config to be merged into the initial config, sent as text. This
        argument is ignored when ``merge_path`` is set.

    merge_path
        Absolute or remote path from where to load the merge configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``initial_path`` or ``merge_path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.merge_diff initial_path=salt://path/to/running.cfg merge_path=salt://path/to/merge.cfg
    """
    if initial_path:
        initial_config = __salt__["cp.get_file_str"](initial_path, saltenv=saltenv)
    candidate_config = merge_text(
        initial_config=initial_config,
        merge_config=merge_config,
        merge_path=merge_path,
        saltenv=saltenv,
    )
    clean_running_dict = tree(config=initial_config)
    clean_running = _print_config_text(clean_running_dict)
    return _get_diff_text(clean_running, candidate_config)


def diff_tree(
    candidate_config=None,
    candidate_path=None,
    running_config=None,
    running_path=None,
    saltenv="base",
):
    """
    Return the diff, as Python dictionary, between the candidate and the running
    configuration.

    candidate_config
        The candidate configuration sent as text. This argument is ignored when
        ``candidate_path`` is set.

    candidate_path
        Absolute or remote path from where to load the candidate configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    running_config
        The running configuration sent as text. This argument is ignored when
        ``running_path`` is set.

    running_path
        Absolute or remote path from where to load the running configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``candidate_path`` or ``running_path`` is not a
        ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.diff_tree candidate_path=salt://path/to/candidate.cfg running_path=salt://path/to/running.cfg
    """
    candidate_tree = tree(config=candidate_config, path=candidate_path, saltenv=saltenv)
    running_tree = tree(config=running_config, path=running_path, saltenv=saltenv)
    return salt.utils.dictdiffer.deep_diff(running_tree, candidate_tree)


def diff_text(
    candidate_config=None,
    candidate_path=None,
    running_config=None,
    running_path=None,
    saltenv="base",
):
    """
    Return the diff, as text, between the candidate and the running config.

    candidate_config
        The candidate configuration sent as text. This argument is ignored when
        ``candidate_path`` is set.

    candidate_path
        Absolute or remote path from where to load the candidate configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    running_config
        The running configuration sent as text. This argument is ignored when
        ``running_path`` is set.

    running_path
        Absolute or remote path from where to load the running configuration
        text. This argument allows any URI supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`), e.g., ``salt://``,
        ``https://``, ``s3://``, ``ftp:/``, etc.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``candidate_path`` or ``running_path`` is not a
        ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' iosconfig.diff_text candidate_path=salt://path/to/candidate.cfg running_path=salt://path/to/running.cfg
    """
    candidate_text = clean(
        config=candidate_config, path=candidate_path, saltenv=saltenv
    )
    running_text = clean(config=running_config, path=running_path, saltenv=saltenv)
    return _get_diff_text(running_text, candidate_text)



"""
Installation and activation of windows licenses
===============================================

Install and activate windows licenses

.. code-block:: yaml

    XXXXX-XXXXX-XXXXX-XXXXX-XXXXX:
      license.activate
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "license"


def __virtual__():
    """
    Only work on Windows
    """
    if salt.utils.platform.is_windows():
        return __virtualname__
    return (False, "Only Windows OS supported")


def activate(name):
    """
    Install and activate the given product key

    name
        The 5x5 product key given to you by Microsoft

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    product_key = name

    license_info = __salt__["license.info"]()
    licensed = False
    key_match = False
    if license_info is not None:
        licensed = license_info["licensed"]
        key_match = license_info["partial_key"] in product_key

    if not key_match:
        out = __salt__["license.install"](product_key)
        licensed = False
        if "successfully" not in out:
            ret["result"] = False
            ret["comment"] += "Unable to install the given product key is it valid?"
            return ret
    if not licensed:
        out = __salt__["license.activate"]()
        if "successfully" not in out:
            ret["result"] = False
            ret["comment"] += "Unable to activate the given product key."
            return ret
        ret["comment"] += "Windows is now activated."
    else:
        ret["comment"] += "Windows is already activated."
    return ret



"""
Execution module for `ciscoconfparse <http://www.pennington.net/py/ciscoconfparse/index.html>`_

.. versionadded:: 2019.2.0

This module can be used for basic configuration parsing, audit or validation
for a variety of network platforms having Cisco IOS style configuration (one
space indentation), including: Cisco IOS, Cisco Nexus, Cisco IOS-XR,
Cisco IOS-XR, Cisco ASA, Arista EOS, Brocade, HP Switches, Dell PowerConnect
Switches, or Extreme Networks devices. In newer versions, ``ciscoconfparse``
provides support for brace-delimited configuration style as well, for platforms
such as: Juniper Junos, Palo Alto, or F5 Networks.

See http://www.pennington.net/py/ciscoconfparse/index.html for further details.

:depends: ciscoconfparse

This module depends on the Python library with the same name,
``ciscoconfparse`` - to install execute: ``pip install ciscoconfparse``.
"""

from salt.exceptions import SaltException

# Import Salt modules

try:
    import ciscoconfparse

    HAS_CISCOCONFPARSE = True
except ImportError:
    HAS_CISCOCONFPARSE = False

# ------------------------------------------------------------------------------
# module properties
# ------------------------------------------------------------------------------

__virtualname__ = "ciscoconfparse"

# ------------------------------------------------------------------------------
# property functions
# ------------------------------------------------------------------------------


def __virtual__():
    if HAS_CISCOCONFPARSE:
        return HAS_CISCOCONFPARSE
    else:
        return (False, "Missing dependency ciscoconfparse")


# ------------------------------------------------------------------------------
# helper functions -- will not be exported
# ------------------------------------------------------------------------------


def _get_ccp(config=None, config_path=None, saltenv="base"):
    """ """
    if config_path:
        config = __salt__["cp.get_file_str"](config_path, saltenv=saltenv)
        if config is False:
            raise SaltException(f"{config_path} is not available")
    if isinstance(config, str):
        config = config.splitlines()
    ccp = ciscoconfparse.CiscoConfParse(config)
    return ccp


# ------------------------------------------------------------------------------
# callable functions
# ------------------------------------------------------------------------------


def find_objects(config=None, config_path=None, regex=None, saltenv="base"):
    """
    Return all the line objects that match the expression in the ``regex``
    argument.

    .. warning::
        This function is mostly valuable when invoked from other Salt
        components (i.e., execution modules, states, templates etc.). For CLI
        usage, please consider using
        :py:func:`ciscoconfparse.find_lines <salt.ciscoconfparse_mod.find_lines>`

    config
        The configuration sent as text.

        .. note::
            This argument is ignored when ``config_path`` is specified.

    config_path
        The absolute or remote path to the file with the configuration to be
        parsed. This argument supports the usual Salt filesystem URIs, e.g.,
        ``salt://``, ``https://``, ``ftp://``, ``s3://``, etc.

    regex
        The regular expression to match the lines against.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file. This
        argument is ignored when ``config_path`` is not a ``salt://`` URL.

    Usage example:

    .. code-block:: python

        objects = __salt__['ciscoconfparse.find_objects'](config_path='salt://path/to/config.txt',
                                                          regex='Gigabit')
        for obj in objects:
            print(obj.text)
    """
    ccp = _get_ccp(config=config, config_path=config_path, saltenv=saltenv)
    lines = ccp.find_objects(regex)
    return lines


def find_lines(config=None, config_path=None, regex=None, saltenv="base"):
    """
    Return all the lines (as text) that match the expression in the ``regex``
    argument.

    config
        The configuration sent as text.

        .. note::
            This argument is ignored when ``config_path`` is specified.

    config_path
        The absolute or remote path to the file with the configuration to be
        parsed. This argument supports the usual Salt filesystem URIs, e.g.,
        ``salt://``, ``https://``, ``ftp://``, ``s3://``, etc.

    regex
        The regular expression to match the lines against.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file. This
        argument is ignored when ``config_path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' ciscoconfparse.find_lines config_path=https://bit.ly/2mAdq7z regex='ip address'

    Output example:

    .. code-block:: text

       cisco-ios-router:
            -  ip address dhcp
            -  ip address 172.20.0.1 255.255.255.0
            -  no ip address
    """
    lines = find_objects(
        config=config, config_path=config_path, regex=regex, saltenv=saltenv
    )
    return [line.text for line in lines]


def find_objects_w_child(
    config=None,
    config_path=None,
    parent_regex=None,
    child_regex=None,
    ignore_ws=False,
    saltenv="base",
):
    """
    Parse through the children of all parent lines matching ``parent_regex``,
    and return a list of child objects, which matched the ``child_regex``.

    .. warning::
        This function is mostly valuable when invoked from other Salt
        components (i.e., execution modules, states, templates etc.). For CLI
        usage, please consider using
        :py:func:`ciscoconfparse.find_lines_w_child <salt.ciscoconfparse_mod.find_lines_w_child>`

    config
        The configuration sent as text.

        .. note::
            This argument is ignored when ``config_path`` is specified.

    config_path
        The absolute or remote path to the file with the configuration to be
        parsed. This argument supports the usual Salt filesystem URIs, e.g.,
        ``salt://``, ``https://``, ``ftp://``, ``s3://``, etc.

    parent_regex
        The regular expression to match the parent lines against.

    child_regex
        The regular expression to match the child lines against.

    ignore_ws: ``False``
        Whether to ignore the white spaces.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file. This
        argument is ignored when ``config_path`` is not a ``salt://`` URL.

    Usage example:

    .. code-block:: python

        objects = __salt__['ciscoconfparse.find_objects_w_child'](config_path='https://bit.ly/2mAdq7z',
                                                                  parent_regex='line con',
                                                                  child_regex='stopbits')
        for obj in objects:
            print(obj.text)
    """
    ccp = _get_ccp(config=config, config_path=config_path, saltenv=saltenv)
    lines = ccp.find_objects_w_child(parent_regex, child_regex, ignore_ws=ignore_ws)
    return lines


def find_lines_w_child(
    config=None,
    config_path=None,
    parent_regex=None,
    child_regex=None,
    ignore_ws=False,
    saltenv="base",
):
    r"""
    Return a list of parent lines (as text)  matching the regular expression
    ``parent_regex`` that have children lines matching ``child_regex``.

    config
        The configuration sent as text.

        .. note::
            This argument is ignored when ``config_path`` is specified.

    config_path
        The absolute or remote path to the file with the configuration to be
        parsed. This argument supports the usual Salt filesystem URIs, e.g.,
        ``salt://``, ``https://``, ``ftp://``, ``s3://``, etc.

    parent_regex
        The regular expression to match the parent lines against.

    child_regex
        The regular expression to match the child lines against.

    ignore_ws: ``False``
        Whether to ignore the white spaces.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file. This
        argument is ignored when ``config_path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' ciscoconfparse.find_lines_w_child config_path=https://bit.ly/2mAdq7z parent_line='line con' child_line='stopbits'
        salt '*' ciscoconfparse.find_lines_w_child config_path=https://bit.ly/2uIRxau parent_regex='ge-(.*)' child_regex='unit \d+'
    """
    lines = find_objects_w_child(
        config=config,
        config_path=config_path,
        parent_regex=parent_regex,
        child_regex=child_regex,
        ignore_ws=ignore_ws,
        saltenv=saltenv,
    )
    return [line.text for line in lines]


def find_objects_wo_child(
    config=None,
    config_path=None,
    parent_regex=None,
    child_regex=None,
    ignore_ws=False,
    saltenv="base",
):
    """
    Return a list of parent ``ciscoconfparse.IOSCfgLine`` objects, which matched
    the ``parent_regex`` and whose children did *not* match ``child_regex``.
    Only the parent ``ciscoconfparse.IOSCfgLine`` objects will be returned. For
    simplicity, this method only finds oldest ancestors without immediate
    children that match.

    .. warning::
        This function is mostly valuable when invoked from other Salt
        components (i.e., execution modules, states, templates etc.). For CLI
        usage, please consider using
        :py:func:`ciscoconfparse.find_lines_wo_child <salt.ciscoconfparse_mod.find_lines_wo_child>`

    config
        The configuration sent as text.

        .. note::
            This argument is ignored when ``config_path`` is specified.

    config_path
        The absolute or remote path to the file with the configuration to be
        parsed. This argument supports the usual Salt filesystem URIs, e.g.,
        ``salt://``, ``https://``, ``ftp://``, ``s3://``, etc.

    parent_regex
        The regular expression to match the parent lines against.

    child_regex
        The regular expression to match the child lines against.

    ignore_ws: ``False``
        Whether to ignore the white spaces.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file. This
        argument is ignored when ``config_path`` is not a ``salt://`` URL.

    Usage example:

    .. code-block:: python

        objects = __salt__['ciscoconfparse.find_objects_wo_child'](config_path='https://bit.ly/2mAdq7z',
                                                                   parent_regex='line con',
                                                                   child_regex='stopbits')
        for obj in objects:
            print(obj.text)
    """
    ccp = _get_ccp(config=config, config_path=config_path, saltenv=saltenv)
    lines = ccp.find_objects_wo_child(parent_regex, child_regex, ignore_ws=ignore_ws)
    return lines


def find_lines_wo_child(
    config=None,
    config_path=None,
    parent_regex=None,
    child_regex=None,
    ignore_ws=False,
    saltenv="base",
):
    """
    Return a list of parent ``ciscoconfparse.IOSCfgLine`` lines as text, which
    matched the ``parent_regex`` and whose children did *not* match ``child_regex``.
    Only the parent ``ciscoconfparse.IOSCfgLine`` text lines  will be returned.
    For simplicity, this method only finds oldest ancestors without immediate
    children that match.

    config
        The configuration sent as text.

        .. note::
            This argument is ignored when ``config_path`` is specified.

    config_path
        The absolute or remote path to the file with the configuration to be
        parsed. This argument supports the usual Salt filesystem URIs, e.g.,
        ``salt://``, ``https://``, ``ftp://``, ``s3://``, etc.

    parent_regex
        The regular expression to match the parent lines against.

    child_regex
        The regular expression to match the child lines against.

    ignore_ws: ``False``
        Whether to ignore the white spaces.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file. This
        argument is ignored when ``config_path`` is not a ``salt://`` URL.

    CLI Example:

    .. code-block:: bash

        salt '*' ciscoconfparse.find_lines_wo_child config_path=https://bit.ly/2mAdq7z parent_line='line con' child_line='stopbits'
    """
    lines = find_objects_wo_child(
        config=config,
        config_path=config_path,
        parent_regex=parent_regex,
        child_regex=child_regex,
        ignore_ws=ignore_ws,
        saltenv=saltenv,
    )
    return [line.text for line in lines]


def filter_lines(
    config=None, config_path=None, parent_regex=None, child_regex=None, saltenv="base"
):
    """
    Return a list of detailed matches, for the configuration blocks (parent-child
    relationship) whose parent respects the regular expressions configured via
    the ``parent_regex`` argument, and the child matches the ``child_regex``
    regular expression. The result is a list of dictionaries with the following
    keys:

    - ``match``: a boolean value that tells whether ``child_regex`` matched any
      children lines.
    - ``parent``: the parent line (as text).
    - ``child``: the child line (as text). If no child line matched, this field
      will be ``None``.

    Note that the return list contains the elements that matched the parent
    condition, the ``parent_regex`` regular expression. Therefore, the ``parent``
    field will always have a valid value, while ``match`` and ``child`` may
    default to ``False`` and ``None`` respectively when there is not child match.

    CLI Example:

    .. code-block:: bash

        salt '*' ciscoconfparse.filter_lines config_path=https://bit.ly/2mAdq7z parent_regex='Gigabit' child_regex='shutdown'

    Example output (for the example above):

    .. code-block:: python

        [
            {
                'parent': 'interface GigabitEthernet1',
                'match': False,
                'child': None
            },
            {
                'parent': 'interface GigabitEthernet2',
                'match': True,
                'child': ' shutdown'
            },
            {
                'parent': 'interface GigabitEthernet3',
                'match': True,
                'child': ' shutdown'
            }
        ]
    """
    ret = []
    ccp = _get_ccp(config=config, config_path=config_path, saltenv=saltenv)
    parent_lines = ccp.find_objects(parent_regex)
    for parent_line in parent_lines:
        child_lines = parent_line.re_search_children(child_regex)
        if child_lines:
            for child_line in child_lines:
                ret.append(
                    {
                        "match": True,
                        "parent": parent_line.text,
                        "child": child_line.text,
                    }
                )
        else:
            ret.append({"match": False, "parent": parent_line.text, "child": None})
    return ret
